
from __future__ import print_function
import os
import sys
import glob
import shutil
import logging

import muteria.common.matrices as common_matrices
import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.drivers.testgeneration.base_testcasetool import BaseTestcaseTool
from muteria.drivers.testgeneration.testcases_info import TestcasesInfoObject

ERROR_HANDLER = common_mix.ErrorHandler

class CustomTestcases(BaseTestcaseTool):
    CUSTOM_TEST_TOOLNAME = "custom_devtests"

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        return True
    #~ def installed()

    def __init__(self, *args, **kwargs):
        BaseTestcaseTool.__init__(self, *args, **kwargs)
        self.test_list_storage_file = os.path.join(self.tests_storage_dir, \
                                                            'test_list.json')
        self.wrapper_obj = self.code_builds_factory.repository_manager.\
                                                        get_wrapper_object()
        self.splittest_ext = '.splittest@'
        self.wrapper_test_splitting = (self.wrapper_obj is not None \
                and self.wrapper_obj.get_test_splitting_wrapper() is not None\
                and not self.config.TESTS_ORACLE_TESTS)
    #~ def __init__()

    def get_testcase_info_object(self):
        tc_info_obj = TestcasesInfoObject()
        ERROR_HANDLER.assert_true(os.path.isfile(self.test_list_storage_file),\
                    "No test list file found, did generation occur?", __file__)
        dtl = common_fs.loadJSON(self.test_list_storage_file)
        for tc, targs in dtl:
            tc_info_obj.add_test(tc)
        return tc_info_obj
    #~ def get_testcase_info_object()

    def execute_testcase (self, testcase, exe_path_map, env_vars, \
                                            timeout=None, \
                                            use_recorded_timeout_times=None, \
                                            recalculate_execution_times=False,\
                                            with_output_summary=True, \
                                            hash_outlog=True):
        return self._in_repo_execute_testcase(testcase=testcase, \
                                exe_path_map=exe_path_map, \
                                env_vars=env_vars, \
                                timeout=timeout, \
                                use_recorded_timeout_times=\
                                    use_recorded_timeout_times, \
                                recalculate_execution_times=\
                                    recalculate_execution_times, \
                                with_output_summary=with_output_summary, \
                                hash_outlog=hash_outlog, \
                                copy_exe_to_repo=(self.wrapper_obj is None))
    #~ def execute_testcase()

    def runtests(self, testcases, exe_path_map, env_vars, \
                                stop_on_failure=False, per_test_timeout=None, \
                                use_recorded_timeout_times=None, \
                                recalculate_execution_times=False, \
                                with_output_summary=True, hash_outlog=True, \
                                parallel_count=1):
        """ Override runtests
        """
        return self._in_repo_runtests(testcases=testcases, \
                                exe_path_map=exe_path_map, \
                                env_vars=env_vars, \
                                stop_on_failure=stop_on_failure, \
                                per_test_timeout=per_test_timeout, \
                                use_recorded_timeout_times=\
                                        use_recorded_timeout_times, \
                                recalculate_execution_times=\
                                        recalculate_execution_times, \
                                with_output_summary=with_output_summary, \
                                hash_outlog=hash_outlog, \
                                parallel_count=parallel_count, \
                                copy_exe_to_repo=(self.wrapper_obj is None))
    #~ def runtests()

    def _prepare_executable(self, exe_path_map, env_vars, \
                                                        collect_output=False):
        """ Make sure we have the right executable ready (if needed)
        """
        #self.code_builds_factory.copy_into_repository(exe_path_map)
        
        if self.wrapper_obj is not None:
            self.wrapper_obj.install_wrapper(exe_path_map, collect_output)
        '''else:
            if exe_path_map is not None:
                repo_root_dir = self.code_builds_factory.repository_manager\
                                                    .get_repository_dir_path()
                for rep_rel_name, abs_name in exe_path_map.items():
                    rep_abs_name = os.path.join(repo_root_dir, rep_rel_name)
                    if abs_name is not None and rep_abs_name != abs_name:
                        # copy to repo
                        try:
                            shutil.copy2(abs_name, rep_abs_name)
                        except PermissionError:
                            os.remove(rep_abs_name)
                            shutil.copy2(abs_name, rep_abs_name)'''
    #~ def _prepare_executable()

    def _restore_default_executable(self, exe_path_map, env_vars, \
                                                        collect_output=False):
        """ Restore back the default executable (if needed).
            Useful for test execution that require the executable
            at a specific location.
        """
        #self.code_builds_factory.restore_repository_files(exe_path_map)
        
        if self.wrapper_obj is not None:
            self.wrapper_obj.uninstall_wrapper(exe_path_map)
        else:
            self.code_builds_factory.set_repo_to_build_default()
    #~ def _restore_default_executable()

    def _execute_a_test (self, testcase, exe_path_map, env_vars, \
                                callback_object=None, timeout=None, \
                                                    collect_output=None):
        """ Execute a test given that the executables have been set 
            properly
        """
        if timeout is None:
            timeout = self.config.ONE_TEST_EXECUTION_TIMEOUT

        #logging.debug('TIMEOUT: '+str(timeout))

        rep_mgr = self.code_builds_factory.repository_manager

        if collect_output:
            # possibly existing wrapper data logs are 
            # removed during wrapper install

            collected_output = []
        else:
            collected_output = None
        
        # change test executor to consider the test ID after @
        runner_testcase = testcase
        if self.wrapper_test_splitting:
            tmp = testcase.split(self.splittest_ext)
            if len(tmp) > 1:
                runner_testcase = ''.join(tmp[:-1])
                subtest_id = tmp[-1]
                st_env_vars = self.wrapper_obj.get_test_splitting_wrapper()\
                                        .get_sub_test_id_env_vars(subtest_id)
                if env_vars is None:
                    env_vars = {}
                env_vars.update(st_env_vars)

        pre,verdict,post = rep_mgr.run_dev_test(dev_test_name=runner_testcase,\
                                exe_path_map=exe_path_map, \
                                env_vars=env_vars, \
                                timeout=timeout,\
                                collected_output=(collected_output \
                                    if self.wrapper_obj is None else None),\
                                callback_object=callback_object)
        ERROR_HANDLER.assert_true(\
                            pre == common_mix.GlobalConstants.COMMAND_SUCCESS,\
                                            "before command failed", __file__)
        ERROR_HANDLER.assert_true(\
                        post != common_mix.GlobalConstants.COMMAND_FAILURE,\
                                            "after command failed", __file__)

        # abort is test execution error
        if verdict == common_mix.GlobalConstants.TEST_EXECUTION_ERROR:
            ERROR_HANDLER.assert_true(\
                                self.config.TEST_EXECUTION_ERROR_AS_FAIL, \
                                "Test Execution error in custom_dev_testcase" \
                                    + " for test: "+testcase, __file__)

        # wrapper cleanup
        if self.wrapper_obj is not None:
            if collect_output:
                self.wrapper_obj.collect_output(\
                                    exe_path_map, collected_output, testcase)
            self.wrapper_obj.cleanup_logs(exe_path_map)

        return verdict, collected_output
    #~ def _execute_a_test()

    def _do_generate_tests (self, exe_path_map, code_builds_factory, \
                                                meta_criteria_tool_obj=None, \
                                                                max_time=None):
        dtl = self.code_builds_factory.repository_manager.get_dev_tests_list()
        ERROR_HANDLER.assert_true(dtl is not None, "invalid dev_test_list", \
                                                                    __file__)
        if os.path.isdir(self.tests_storage_dir):
            shutil.rmtree(self.tests_storage_dir)
        os.mkdir(self.tests_storage_dir)

        if self.wrapper_test_splitting:
            # Execute the tests with the counting wrapper
            logging.debug("## Splitting tests with wrapper")
            split_workdir = os.path.join(self.tests_storage_dir)
            new_exe_path_map = self.wrapper_obj.get_test_splitting_wrapper()\
                                    .set_wrapper(split_workdir, exe_path_map)
            sub_test_list = []
            for test in dtl:
                logging.debug('splitting test '+test) #DBG
                self.wrapper_obj.get_test_splitting_wrapper()\
                                            .switch_to_new_test()
                self.execute_testcase(test, new_exe_path_map, {})
                n_subtest, args = \
                                self.wrapper_obj.get_test_splitting_wrapper()\
                                                                .collect_data()
                for i in range(n_subtest):
                    sub_test_list.append([test+self.splittest_ext+str(i), \
                                        args[i] if i < len(args) else None])

            self.wrapper_obj.get_test_splitting_wrapper().cleanup()

            # update dtl with counter tests
            dtl = sub_test_list
        else:
            dtl = [[t, None] for t in dtl]

        common_fs.dumpJSON(dtl, self.test_list_storage_file)
        #ERROR_HANDLER.assert_true(os.path.isfile(self.test_list_storage_file))
    #~ def _do_generate_tests()
#~ class CustomTestcases



import os
import logging
import importlib
import subprocess
import signal
import time

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix
import muteria.common.matrices as common_matrices

ERROR_HANDLER = common_mix.ErrorHandler

class ToolsModulesLoader(object):
    """ Load tools drivers
        Example:
        >>> import muteria.drivers as md
        >>> md.ToolsModulesLoader.get_tools_modules(\
        ...                            md.ToolsModulesLoader.CRITERIA_TOOLS)
        >>> {}
    """

    TESTCASES_TOOLS = "TESTCASES"
    CRITERIA_TOOLS = "CRITERIA"

    # map of directory modules by categories 
    TOOL_CATEGORIES_DIRS = {
        TESTCASES_TOOLS: "testgeneration",
        CRITERIA_TOOLS: "criteria",
    }

    # directory in each category's module that contain language folders, 
    # containing corresponding tools
    COMMON_TOOLS_BY_LANGUAGE_DIR = "tools_by_languages"

    @classmethod
    def get_tools_modules(cls, tool_category):
        '''
        Import all tools modules of the specified tool category into a Dict 
        as following:  {<language>: {<toolname>: <module>}}
        '''
        if tool_category not in cls.TOOL_CATEGORIES_DIRS:
            logging.error("%s: %s" % ("invalid tool_category", tool_category))

        modules = {}
        
        to_skip = ('__pycache__',)

        drivers_pkg_init_py = os.path.abspath(__file__)
        drivers_pkg_path = os.path.dirname(drivers_pkg_init_py)
        pkg_path = os.path.join(drivers_pkg_path, \
                                    cls.TOOL_CATEGORIES_DIRS[tool_category], \
                                    cls.COMMON_TOOLS_BY_LANGUAGE_DIR)
        language_list = []
        for lang in os.listdir(pkg_path):
            if os.path.isdir(os.path.join(pkg_path, lang)) and \
                                                        lang not in to_skip:
                language_list.append(lang.lower())

        for language in language_list:
            if '.' in language:
                ERROR_HANDLER.error_exit("%s (%s). %s" % \
                                    ("Malformed language name", language, \
                        "Language (package) name must not contain dot('.')"), \
                                                                    __file__)
            language_pkg_path = os.path.join(pkg_path, language)
            toolname_list = []
            for fd in os.listdir(language_pkg_path):
                if os.path.isdir(os.path.join(language_pkg_path, fd)) and \
                                                            fd not in to_skip:
                    toolname_list.append(fd)

            for toolname in toolname_list:
                if '.' in toolname:
                    ERROR_HANDLER.error_exit("%s (%s). %s" % \
                                    ("Malformed tool name", toolname, \
                        "Toolname (package) name must not contain dot('.')"),\
                                                                    __file__)
                    
                load_string = ".".join(['muteria', 'drivers', \
                                    cls.TOOL_CATEGORIES_DIRS[tool_category],\
                                    cls.COMMON_TOOLS_BY_LANGUAGE_DIR,\
                                                        language, toolname])

                # import the mutation tool 
                tool_pkg = importlib.import_module(load_string)

                # Add into dict
                if language not in modules:
                    modules[language] = {}
                if toolname in modules[language]:
                    logging.error("%s %s %s %s " % ("tool", toolname, \
                        "appearing multiple times for same language", language))
                    ERROR_HANDLER.error_exit()
                modules[language][toolname] = tool_pkg

        return modules
#~ class ToolsModulesLoader

class RepoFileToCustomMap(dict):
    """ This class define a mapping between a repository file and its 
    custom coresponding file. This is usefule to let the testcase tool
    know which file to replace during test execution.
    Just specify the files to change. If none is specified, the default 
    files in the repository are used.
    """

    def __init__(self):
        pass
    #~ def __init__()

    def add_map(self, repo_val, custom_val):
        ERROR_HANDLER.assert_true(repo_val not in self, \
                        "adding map failed: key {} already present.".format( \
                            repo_val), __file__)
        self[repo_val] = custom_val
    #~ def add_map()

    def update_custom_value(self, repo_val, custom_val):
        ERROR_HANDLER.assert_true(repo_val in self, \
                        "updating map failed: key {} not present.".format( \
                            repo_val), __file__)
        self[repo_val] = custom_val
    #~ def update_custom_value()

    def delete_map(self, repo_val):
        ERROR_HANDLER.assert_true(repo_val in self, \
                        "deleting map failed: key {} not present.".format( \
                            repo_val), __file__)
        del self[repo_val]
    #~ def get_custom_value()

    def get_custom_value(self, repo_val):
        ERROR_HANDLER.assert_true(repo_val in self, \
                        "getting map failed: key {} not present.".format( \
                            repo_val), __file__)
        return self[repo_val]
    #~ def get_custom_value()

    def write_to_file(self, filepath):
        common_fs.dumpJSON(self, filepath)
    #~ def write_to_file()

    def read_from_file(self, filepath):
        obj = common_fs.loadJSON(filepath)
        for k in obj:
            self.add_map(k, obj[k])
    #~ def write_to_file()
        
#~ class RepoFileToCustomMap

class DriversUtils(object):

    ################### Meta to non meta and vice versa ####################
    
    @classmethod
    def make_meta_element(cls, element, toolalias):
        return ":".join([toolalias, element])
    #~ def make_meta_element()

    @classmethod
    def reverse_meta_element(cls, meta_element):
        parts = meta_element.split(':', 1)
        assert len(parts) >= 2, "invalid meta mutant"
        toolalias, element = parts
        return toolalias, element
    #~ def reverse_meta_element()

    ############################ Subprocess Stuffs #########################
 
    # XXX sync timed out signal in retcode with used in execute_and_get...
    #EXEC_TIMED_OUT_RET_CODE = (-signal.SIGKILL, -signal.SIGTERM,) # < python 3
    EXEC_TIMED_OUT_RET_CODE = (-signal.SIGKILL.value, -signal.SIGTERM.value,)
    EXEC_SEGFAULT_OUT_RET_CODE = (-11,)

    @classmethod
    def check_tool(cls, prog, args_list=[], expected_exit_codes=[0]):
        try:
            p = subprocess.Popen([prog]+args_list, env=os.environ, \
                                             #close_fds=True, \
                                            stderr=subprocess.DEVNULL,\
                                            stdout=subprocess.DEVNULL)
            retcode = p.wait()
            if retcode not in expected_exit_codes:
                ERROR_HANDLER.error_exit("Program {} {}.".format(prog,\
                                        ': check is problematic'), __file__)
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                logging.info("{} not installed".format(prog))
                return False
            else:
                ERROR_HANDLER.error_exit("Something went wrong", __file__)
        return True
    #~ def check_tool()

    @classmethod
    def execute_and_get_retcode_out_err(cls, prog, args_list=[], \
                            env=None, timeout=None, timeout_grace_period=5, \
                            out_on=True, err_on=True, merge_err_to_out=True, \
                            cwd=None):
        #print(prog, args_list, env is None, timeout, out_on, err_on, merge_err_to_out)
        tmp_env = os.environ if env is None else env
        out = subprocess.PIPE if out_on else subprocess.DEVNULL
        if err_on:
            err = subprocess.STDOUT if merge_err_to_out else subprocess.PIPE
        else:
            err = subprocess.DEVNULL
        # use setsid to kill the process group
        p = subprocess.Popen([prog]+args_list, env=tmp_env, cwd=cwd, \
                                                            #close_fds=True, \
                                                        stderr=err,\
                                                        stdout=out, 
                                                        preexec_fn=os.setsid)
        try:
            stdout, stderr = p.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            #p.terminate() # TODO: Chose the signal to send
            group_id = os.getpgid(p.pid)
            #os.killpg(p.pid, signal.SIGTERM)
            os.killpg(group_id, signal.SIGTERM)
            #p.send_signal(signal.SIGINT) # TODO: Chose the signal to send
            # give timeout_grace_period seconds to stop
            stopped = False
            for _ in range(timeout_grace_period):
                if p.poll() is None:
                    time.sleep(1)
                else:
                    stopped = True
                    break
            if not stopped:
                #os.killpg(p.pid, signal.SIGKILL)
                os.killpg(group_id, signal.SIGKILL)
                p.kill() # TODO: Chose the signal to send
            stdout, stderr = p.communicate()
        if stdout is not None:
            stdout = stdout.decode('UTF-8', 'backslashreplace')
        if stderr is not None:
            stderr = stderr.decode('UTF-8', 'backslashreplace')
        #retcode = p.wait()
        retcode = p.poll()
        ERROR_HANDLER.assert_true(retcode is not None, \
                                "child process still running after timeout")
        return retcode, stdout, stderr
    #~ def execute_and_get_retcode_out_err()

    ############################### Misc ###################################

    @classmethod
    def update_matrix_to_cover_when_difference(cls, \
                                target_matrix_file, target_outdata_file, \
                                comparing_vector_file, comparing_outdata_file):
        ERROR_HANDLER.assert_true(target_matrix_file is not None \
                                and comparing_vector_file is not None, \
                                "target or comparing matrix is None", __file__)

        target_matrix = common_matrices.ExecutionMatrix(\
                                                filename=target_matrix_file)
        comparing_vector = common_matrices.ExecutionMatrix(\
                                                filename=comparing_vector_file)

        ERROR_HANDLER.assert_true(len(\
                            set(target_matrix.get_nonkey_colname_list()) - \
                            set(comparing_vector.get_nonkey_colname_list()) \
                                     ) == 0, "Mismatch of columns", __file__)

        # Get uncertain
        target_uncertain_cols_dict = \
                                target_matrix.query_uncertain_columns_of_rows()
        vector_uncertain_cols = \
                            comparing_vector.query_uncertain_columns_of_rows()
        vector_uncertain_cols = \
                    set(vector_uncertain_cols[list(vector_uncertain_cols)[0]])

        # Check if outdata and proceed accordingly
        if target_outdata_file is not None and \
                                            comparing_outdata_file is not None:
            # outdata are set use outdata values to decide of differences
            target_outdata = common_matrices.OutputLogData(\
                                                filename=target_outdata_file)
            vector_outdata = common_matrices.OutputLogData(\
                                            filename=comparing_outdata_file)

            _, vector_outdata_uniq = list(\
                                vector_outdata.get_zip_objective_and_data())[0]

            ## Compare using output
            timedout_key = common_matrices.OutputLogData.TIMEDOUT
            key_to_diffs = {}
            for key, key_data in target_outdata.get_zip_objective_and_data():
                intersect = set(vector_outdata_uniq) & set(key_data)
                ERROR_HANDLER.assert_true(set(key_data) == intersect, \
                            "The elements in target must all be in vector",\
                                                                    __file__)
                key_to_diffs[key] = set()
                for elem in intersect:
                    ol_equiv = common_matrices.OutputLogData.outlogdata_equiv(\
                                    key_data[elem], vector_outdata_uniq[elem])
                    if not ol_equiv:
                        key_to_diffs[key].add(elem)
        else:
            # outdata is not set use difference of matrices
            ## obtain the active cols for each rows
            target_active_cols_dict = \
                                target_matrix.query_active_columns_of_rows()
            
            vector_active_cols = \
                                comparing_vector.query_active_columns_of_rows()
            vector_active_cols = \
                        set(vector_active_cols[list(vector_active_cols)[0]])

            ## for each row of target matrix, get diff
            target_matrix_allcol = set(target_matrix.get_nonkey_colname_list())
            key_to_diffs = {}
            for row_key, row_active in target_active_cols_dict.items():
                key_to_diffs[row_key] = (set(row_active) - vector_active_cols)\
                                    | (vector_active_cols - set(row_active))
                key_to_diffs[row_key] &= target_matrix_allcol

        # Update matrix based on diff
        for key, diffs in key_to_diffs.items():
            # gather the uncertain and set to uncertain
            uncertain = vector_uncertain_cols | \
                                        set(target_uncertain_cols_dict[key])
            set_uncertain = uncertain & diffs
            diffs -= uncertain

            values = {}
            for col in set_uncertain:
                values[col] = target_matrix.getUncertainCellDefaultVal()
            for col in diffs:
                values[col] = target_matrix.getActiveCellDefaultVal()

            if len(values) > 0:
                target_matrix.update_cells(key, values)

        target_matrix.serialize()
    #~ def update_matrix_to_cover_when_difference()
#~class DriversUtils()



import os
import logging
import shutil

from muteria.common.mix import GlobalConstants
from muteria.drivers import DriversUtils

from muteria.drivers.testgeneration.custom_dev_testcase.system_wrappers \
                import TEST_FILE_NAME_ENV_VAR, TEST_EXECUTION_TIMEOUT_ENV_VAR

def system_test_runner(prog, args_list, test_filename, repo_root_dir, \
                            exe_path_map=None, env_vars=None, timeout=None, \
                            collected_output=None, using_wrapper=False, \
                            dbg_log_execution_out=False):
    try:
        tmp_env = os.environ.copy()
        if env_vars is not None:
            #for e, v in env_vars.items():
            #    tmp_env[e] = v
            tmp_env.update(env_vars)
        if test_filename is not None:
            tmp_env[TEST_FILE_NAME_ENV_VAR] = test_filename
        if using_wrapper and timeout is not None:
            tmp_env[TEST_EXECUTION_TIMEOUT_ENV_VAR] = str(timeout)
            timeout = None

        if collected_output is None:
            retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=prog, args_list=args_list, env=tmp_env,\
                                timeout=timeout, out_on=dbg_log_execution_out,\
                                err_on=dbg_log_execution_out, \
                                merge_err_to_out=True)
        else:
            retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=prog, args_list=args_list, env=tmp_env,\
                                timeout=timeout, merge_err_to_out=True)
            collected_output.append(retcode)
            collected_output.append(out)

        if dbg_log_execution_out:
            logging.debug("(DBG - Test Output):\n"+out)
    except (ValueError, OSError) as e:
        # ERROR
        # TODO: 
        return GlobalConstants.TEST_EXECUTION_ERROR
    
    # Parse the result
    hasfail = False
    hasfail |= (retcode != 0)

    return GlobalConstants.FAIL_TEST_VERDICT if hasfail else \
                                            GlobalConstants.PASS_TEST_VERDICT
#~ def dev_test_runner()



#
# [LICENCE]
#
""" Base and Default test execution optimizer classes

    Each file that implement a optimizer must have the class 
    `TestExecutionOptimizer` that implements the base class bellow
"""

from __future__ import print_function
import os
import sys
import logging
import abc

import muteria.common.mix as common_mix

ERROR_HANDLER = common_mix.ErrorHandler

class BaseTestExecutionOptimizer(abc.ABC):
    '''
    '''

    def __init__(self, config, explorer, **kwargs):
        # Set Direct Arguments Variables
        self.config = config
        self.explorer = explorer

        self.test_ordered_list = None
        self.pointer = 0

        # by default not locked (can reset)
        self.reset_disabled = False
    #~ def __init__()

    def has_next_test (self):
        return self.pointer < len(self.test_ordered_list)
    #~ def has_next_test()

    def get_next_test (self):
        ERROR_HANDLER.assert_true(self.test_ordered_list is not None, \
                                            'uninitialized test_order_list')
        if self.has_next_test():
            ret = self.test_ordered_list[self.pointer]
            self.pointer += 1
        else:
            ret = None
        return ret
    #~ def get_next_test()

    def select_tests(self, proportion_number, is_proportion=True):
        """ Apply test Selection
        """
        ERROR_HANDLER.assert_true(self.test_ordered_list is not None, \
                                            'uninitialized test_order_list')
        ERROR_HANDLER.assert_true(proportion_number > 0, \
                                        'proportion_number must be positive')
        if is_proportion:
            ERROR_HANDLER.assert_true(proportion_number <= 100, \
                                        'proportion must be <= 100')
            proportion_number = round(len(self.test_ordered_list) * \
                                                    proportion_number / 100.0)
            proportion_number = max(1, int(proportion_number))
        return self.test_ordered_list[:proportion_number]
    #~ def select_tests()

    def feedback (self, test_case, verdict, **kwargs):
        """ Possibly get feedback from past executions
            Override this if needed
        """
    #~ def feedback()

    #######################################################################
    ##################### Methods to implement ############################
    #######################################################################

    @classmethod
    @abc.abstractclassmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def installed()

    @abc.abstractmethod
    def reset (self, toolalias, test_list, disable_reset=False, **kwargs):
        """ Reset the optimizer
            :param disable_reset: disable reset after this call to reset
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def reset()
#~ class BaseTestExecutionOptimizer



from __future__ import print_function

import os

import muteria.common.mix as common_mix

from muteria.drivers.testgeneration.custom_dev_testcase.system_wrappers.\
                                base_wrapper_setup import BaseSystemWrapper, \
                                                BaseSystemTestSplittingWrapper

ERROR_HANDLER = common_mix.ErrorHandler

class SystemTestSplittingWrapper(BaseSystemTestSplittingWrapper):
    def set_wrapper(self, workdir, exe_path_map):
        """ Return the new exe path map
        """
        ERROR_HANDLER.assert_true(os.path.isdir(workdir), "workdir missing", \
                                                                    __file__)
        # For system tests there might be different users
        # Make sure that the next files are accessible by anyone
        os.chmod(workdir, 0o777)

        self.testsplit_wrapper_file = os.path.join(workdir, \
                                                'test_split_wrapper.sh')
        self.counting_file = os.path.join(workdir, \
                                                    'test_split_counter')
        self.splittest_args = os.path.join(workdir, \
                                                    'splittests_args')
        with open(self.testsplit_wrapper_file, 'w') as f:
            f.write("#! /bin/bash\n")
            f.write('set -u\n')
            f.write("count=$(/bin/cat {})\n".format(self.counting_file))
            f.write("count=$(($count + 1))\n")
            f.write("/bin/echo $count > {}\n".format(self.counting_file))
            f.write('/bin/echo " ${{@:1}}" >> {}\n'.format(\
                                                        self.splittest_args))
            f.write(\
                '$MUTERIA_SYSTEM_WRAPPER_W_DEFAULT_TOOL_ABS_NAME "${@:1}"\n')
        os.chmod(self.testsplit_wrapper_file, 0o775)

        new_exe_path_map = dict(exe_path_map)
        ERROR_HANDLER.assert_true(len(new_exe_path_map) == 1, \
                                "Only single executable is supported"
                                + " for wrapper test splitting", __file__)
        for k in new_exe_path_map:
            new_exe_path_map[k] = self.testsplit_wrapper_file
        return new_exe_path_map
    #~ def set_wrapper()

    def switch_to_new_test(self):
        """ reset the counters
        """
        with open(self.counting_file, 'w') as f:
            f.write('-1\n')
        if os.path.isfile(self.splittest_args):
            os.remove(self.splittest_args)
    #~ def switch_to_new_test()

    def collect_data(self):
        """ get number of sub tests and args
        """
        ERROR_HANDLER.assert_true(os.path.isfile(self.splittest_args), \
                            "No args file ({}) during wrapper test split."\
                                .format(self.splittest_args) + \
                        "Does the test have permission to write in repo dir?",\
                                                                     __file__)
        with open(self.counting_file) as f:
            max_id = int(f.read())
            n_subtest = max_id + 1
        try:
            with open(self.splittest_args) as f:
                args = f.read().splitlines()
        except UnicodeDecodeError:
            with open(self.splittest_args, encoding='ISO-8859-1') as f:
                #encode('utf-8', 'backslashreplace')
                args = f.read().splitlines()
        return n_subtest, args
    #~ def collect_data()

    def cleanup(self):
        """ Clean the temps
        """
        for fn in [self.testsplit_wrapper_file, self.counting_file, \
                                                        self.splittest_args]:
            if os.path.isfile(fn):
                os.remove(fn)
    #~ def cleanup()
#~ class SystemTestSplittingWrapper


class SystemWrapper(BaseSystemWrapper):

    wrapper_template_filename = "MUTERIA_WRAPPER.sh.in"

    def get_dev_null(self):
        return "/dev/null"
    #~ def get_dev_null()

    def _get_wrapper_template_string(self):
        wrapper_template_file = os.path.join(os.path.dirname(__file__), \
                                                self.wrapper_template_filename)
        with open(wrapper_template_file) as template:
            return template.read()
    #~ def _get_wrapper_template_string()

    def _get_timedout_codes(self):
        # Timeout codes for GNU timeout
        return (124, 137)
    #~ def _get_timedout_codes()

    def get_test_splitting_wrapper_class(self):
        return SystemTestSplittingWrapper
    #~ def get_sub_test_id_env_vars()
#~ class SystemWrapper


""" This Module implement some utility mathods used throughout the project.
        - The function `confirm_execution` is usefule to request user
            confirmation before proceeding the execution of a task that 
            is considered important to verify user certainty in proceeding.
            An example is the deletion of the data directory...
        - The class `ErrorHandler` define function that are called to 
            terminate the execution gracefully and informativelly 
            upon error. 
"""

from __future__ import print_function

import sys
import shutil
import logging
import inspect
import traceback

import enum

def confirm_execution(question):
    """
    Ask user to enter Y or N (case-insensitive).

    :return: True if the answer is Y.
    :rtype: bool
    """
    reading_func = input
    if sys.version_info.major < 3:
        reading_func = eval('raw_input')
    answer = ""
    while answer not in ["y", "n"]:
        answer = reading_func("%s %s" % (question, "[Y/N] ")).lower()
    return answer == "y"
#~ def confirm_execution()

class ErrorHandler(object):
    repos_dir_manager = None
    # Make sure that there is no infinite recursive call to error_exit
    # if the function `RepositoryManager.revert_repository` make a call
    # to `error_exit`
    error_exit_revert_repo_called = False
    
    def __init__(self):
        pass

    @classmethod
    def set_corresponding_repos_manager(cls, repos_dir_manager):
        cls.assert_true (cls.repos_dir_manager is None, \
                            err_string="the repo dir manager is already set", \
                            call_location=__file__)
        cls.repos_dir_manager = repos_dir_manager
    #~ def set_corresponding_repos_manager()

    @classmethod
    def error_exit(cls, err_string=None, call_location=None, error_code=1, \
                                        ask_revert=True, from_assert=False):
        #traceback.print_stack()
        logging.error('\n'+''.join(traceback.format_stack()))
        if call_location is not None:
            logging.error("# Error happened in location {}".format(\
                                                            call_location))
        if from_assert:
            last_function = inspect.stack()[2][3]
        else:
            last_function = inspect.stack()[1][3]
        logging.error("# Error happened in function %s" % last_function)
                                                        
        if err_string:
            logging.error('')
            logging.error(" (msg) "+err_string)
            logging.error('')
        if ask_revert and cls.repos_dir_manager is not None:
            if cls.error_exit_revert_repo_called:
                logging.error("@ post error: Failed to revert repository."
                                "DO IT MANUALLY!!")
            elif confirm_execution("Do you want to revert repository files?"):
                logging.info("@ post error: Reverting repository files")
                cls.error_exit_revert_repo_called = True
                cls.repos_dir_manager.revert_repository(as_initial=False)
        else:
            logging.info("@ post error: Manually revert the repository files")
        logging.info("# Exiting with code %d" % error_code)
        sys.exit(error_code)
    #~ def error_exit()

    @classmethod
    def assert_true(cls, condition, err_string=None, \
                                call_location=None, ask_revert=True):
        '''
        Call this function with the parameter *__file__*
        as location_called_from
        '''
        if not condition:
            cls.error_exit(err_string=err_string, \
                            call_location=call_location, \
                            ask_revert=ask_revert, from_assert=True)
    #~ def error_exit_file()
#~ class ErrorHandler

class GlobalConstants(object):
    UNCERTAIN_TEST_VERDICT = None

    PASS_TEST_VERDICT = 0
    
    FAIL_TEST_VERDICT = 1

    TEST_EXECUTION_ERROR = -1
    
    ELEMENT_UNCERTAIN_VERDICT = None

    ELEMENT_COVERED_VERDICT = 1
    
    ELEMENT_NOTCOVERED_VERDICT = 0

    COMMAND_UNCERTAIN = None

    COMMAND_SUCCESS = 0

    COMMAND_FAILURE = 1
#~ class GlobalConstants

@enum.unique
class EnumAutoName(enum.Enum):
    
    """ This is a base class that can be used to define enums, 
        Extend this class to generate an enum that will have the methods
        defined here.
    """
        
        # This function do not have 'self'
    #def _generate_next_value_(name, start, count, last_values):
        #return name

    def get_str(self):
        """ get the string representation of the enum field.
            Example:
            >>> class MyEnum(EnumAutoName):
            ...     FIELD1 = "abc"
            ...     FIELD2 = 3
            ...
            >>> my = MyEnum
            >>> my.FIELD1.get_str()
            'FIELD1'
        """
        return self.name
    #~ def get_str():

    def get_field_value(self):
        """ get the value of the enum field.
            Example:
            >>> class MyEnum(EnumAutoName):
            ...     FIELD1 = "abc"
            ...     FIELD2 = 3
            ...
            >>> my = MyEnum
            >>> my.FIELD1.get_field_value()
            'abc'
        """
        return self.value
    #~ def get_field_value():

    @classmethod
    def has_element_named(cls, e_name):
        """ Check that the string e_name represent the name of a field
            of the enum.
            Example:
            >>> class MyEnum(EnumAutoName):
            ...     FIELD1 = "abc"
            ...     FIELD2 = 3
            ...
            >>> MyEnum.has_element_named("FIELD1")
            True
            >>> MyEnum.has_element_named("xyz")
            False
        """
        return e_name in cls.__members__
    #~ def has_element_named()
    
    @classmethod
    def is_valid(cls, elem):
        """ Check that the object elem is a field of the enum
            Example:
            >>> class MyEnum(EnumAutoName):
            ...     FIELD1 = "abc"
            ...     FIELD2 = 3
            ...
            >>> MyEnum.is_valid(MyEnum.FIELD1)
            True
            >>> MyEnum.is_valid("FIELD1")
            False
        """
        return elem in cls
    #~ def is_valid()
#~ class EnumAutoName




from __future__ import print_function

import sys
import os
import logging
import shutil

import muteria.common.mix as common_mix

import muteria.repositoryandcode.codes_convert_support as ccs
from muteria.repositoryandcode.callback_object import DefaultCallbackObject
from muteria.drivers import DriversUtils

ERROR_HANDLER = common_mix.ErrorHandler

__all__ = ['FromC', 'FromCpp']

class FromC(ccs.BaseCodeFormatConverter):
    def __init__(self):
        self.src_formats = [
            ccs.CodeFormats.C_SOURCE,
            ccs.CodeFormats.C_PREPROCESSED_SOURCE
        ]
        self.dest_formats = [
            ccs.CodeFormats.C_PREPROCESSED_SOURCE,
            ccs.CodeFormats.LLVM_BITCODE,
            ccs.CodeFormats.OBJECT_FILE,
            ccs.CodeFormats.NATIVE_CODE,
        ]
    #~ def __init__()

    def convert_code(self, src_fmt, dest_fmt, file_src_dest_map, \
                                                repository_manager, **kwargs):
        # TODO: add can_fail parameter, in kwarg, for case like mutant
        # compilatioon that can fail but should not terminate execution
        # but return a specific value
        ERROR_HANDLER.assert_true(src_fmt in self.src_formats, \
                                    "Unsupported src format", __file__)

        # post build callbacks
        class CopyCallbackObject(DefaultCallbackObject):
            def after_command(self):
                if self.op_retval == \
                                    common_mix.GlobalConstants.COMMAND_FAILURE:
                    return common_mix.GlobalConstants.COMMAND_FAILURE
                for sf, df in list(file_src_dest_map.items()):
                    abs_sf = repository_manager.repo_abs_path(sf)
                    if not os.path.isfile(abs_sf):
                        ERROR_HANDLER.error_exit(\
                                "an expected file missing after build: "+\
                                                            abs_sf, __file__)
                    if df is not None:
                        shutil.copy2(abs_sf, df)
                return None
            #~ def after_command()
        #~ class CopyCallbackObject

        # Should not have callback_object and file_src_dest_map at the
        # same time
        callbak_obj_key = 'callback_object'
        if callbak_obj_key in kwargs:
            ERROR_HANDLER.assert_true(file_src_dest_map is None,\
                            "file_src_dest_map must be None "+ \
                            "if callback_object is passed", __file__)
        elif file_src_dest_map is not None and len(file_src_dest_map) > 0:
            kwargs[callbak_obj_key] = CopyCallbackObject()
        else:
            kwargs[callbak_obj_key] = None

        # Actual Processing
        if (dest_fmt == ccs.CodeFormats.C_PREPROCESSED_SOURCE):
            if (src_fmt == ccs.CodeFormats.C_SOURCE):
                ERROR_HANDLER.error_exit("Must Implement1", __file__)
            else:
                for src, dest in list(file_src_dest_map.items()):
                    shutil.copy2(src, dest)
        if (dest_fmt == ccs.CodeFormats.LLVM_BITCODE):
            # XXX: This build overrides passed clean_tmp and reconfigure
            # also overrides Compiler if wllvm if found
            # and does not use the callback object
            # and require file_src_dest_map to have the place to store
            # generated .bc by specifying the corresponding native file.
            # EX: {x.c: /path/to/main} passed to have /path/to/main.bc
            # generated

            spec_compiler = kwargs['compiler'] if 'compiler' in kwargs \
                                                                    else None
            # special kwargs
            spec_llvm_compiler_path = None 
            if 'llvm_compiler_path' in kwargs:
                spec_llvm_compiler_path = kwargs['llvm_compiler_path']
                del kwargs['llvm_compiler_path']

            # If an llvm compiler is specified, 
            # use it instead of the WLLVM default
            if spec_compiler is not None:
                if 'LLVM_COMPILER' not in os.environ:
                    bak_llvm_compiler = None
                else:
                    bak_llvm_compiler = os.environ['LLVM_COMPILER']
                os.environ['LLVM_COMPILER'] = spec_compiler
            else:
                if 'LLVM_COMPILER' not in os.environ:
                    # default to clang
                    os.environ['LLVM_COMPILER'] = 'clang'

            if spec_llvm_compiler_path is not None:
                if 'LLVM_COMPILER_PATH' not in os.environ:
                    bak_llvm_compiler_path = None
                else:
                    bak_llvm_compiler_path = os.environ['LLVM_COMPILER_PATH']
                os.environ['LLVM_COMPILER_PATH'] = spec_llvm_compiler_path

            #1. Ensure wllvm is installed (For now use default llvm compiler)
            has_wllvm = DriversUtils.check_tool('wllvm', ['--version'])
            ERROR_HANDLER.assert_true(has_wllvm, 'wllvm not found '\
                                                '(To install please visit '\
                            'https://github.com/travitch/whole-program-llvm)', 
                                                                    __file__)

            # tmp['LLVM_COMPILER_PATH'] = ...
            kwargs['compiler'] = 'wllvm'
            kwargs['clean_tmp'] = True
            kwargs['reconfigure'] = True

            # Normal build followed by executable copying
            pre_ret, ret, post_ret = repository_manager.build_code(**kwargs)
            ERROR_HANDLER.assert_true(\
                    ret != common_mix.GlobalConstants.COMMAND_FAILURE and\
                    pre_ret != common_mix.GlobalConstants.COMMAND_FAILURE and\
                    post_ret != common_mix.GlobalConstants.COMMAND_FAILURE,\
                                        "Build LLVM bitcode failed!", __file__)

            # extract bitcode from copied executables and remove non bitcode
            # TODO: make this happen as post callback of compilation with wllvm
            if file_src_dest_map is not None:
                for src, dest in list(file_src_dest_map.items()):
                    ret, out, err = \
                            DriversUtils.execute_and_get_retcode_out_err( \
                                                        "extract-bc", [dest])
                    ERROR_HANDLER.assert_true(ret == 0, \
                                        '{}. \n# OUT: {}\n# ERR: {}'.format(\
                                    'extract-bc failed', out, err), __file__)
                    os.remove(dest)

            if spec_compiler is not None:
                if bak_llvm_compiler is not None:
                    os.environ['LLVM_COMPILER'] = bak_llvm_compiler
                else:
                    del os.environ['LLVM_COMPILER']
            if spec_llvm_compiler_path is not None:
                if bak_llvm_compiler_path is not None:
                    os.environ['LLVM_COMPILER_PATH'] = bak_llvm_compiler_path
                else:
                    del os.environ['LLVM_COMPILER_PATH']

            # Clean build
            kwargs['compiler'] = None
            pre_ret, ret, post_ret = repository_manager.build_code(**kwargs)

        if (dest_fmt == ccs.CodeFormats.OBJECT_FILE):
            ERROR_HANDLER.error_exit("Must Implement3", __file__)
        if (dest_fmt == ccs.CodeFormats.NATIVE_CODE):
            pre_ret, ret, post_ret = repository_manager.build_code(**kwargs)
        return pre_ret, ret, post_ret
    #~ def convert_code()

    def get_source_formats(self):
        return self.src_formats
    #~ def get_source_formats()

    def get_destination_formats_for(self, src_fmt):
        return self.dest_formats
    #~ def get_destination_formats()
#~ class FromC

class FromCpp(ccs.BaseCodeFormatConverter):
    def __init__(self):
        self.src_formats = [
            ccs.CodeFormats.CPP_SOURCE,
            ccs.CodeFormats.CPP_PREPROCESSED_SOURCE
        ]
        self.dest_formats = [
            ccs.CodeFormats.CPP_PREPROCESSED_SOURCE,
            ccs.CodeFormats.LLVM_BITCODE,
            ccs.CodeFormats.OBJECT_FILE,
            ccs.CodeFormats.NATIVE_CODE,
        ]
    #~ def __init__()

    def convert_code(self, src_fmt, dest_fmt, file_src_dest_map, \
                                                repository_manager, **kwargs):
        ERROR_HANDLER.error_exit("Must Implement", __file__)
    #~ def identity_function()

    def get_source_formats(self):
        return self.src_formats
    #~ def get_source_formats()

    def get_destination_formats_for(self, src_fmt):
        return self.dest_formats
    #~ def get_destination_formats()
#~ class FromCpp




from __future__ import print_function

import os
import sys
import logging
import shutil
import abc

import muteria.common.mix as common_mix

from muteria.repositoryandcode.callback_object import DefaultCallbackObject

ERROR_HANDLER = common_mix.ErrorHandler

class CodeFormats(common_mix.EnumAutoName):
    NATIVE_CODE = "NATIVE_CODE"
    OBJECT_FILE = "OBJECT_FILE"
    ASSEMBLY_CODE = "ASSEMBLY_CODE"

    LLVM_BITCODE = "LLVM_BITCODE"

    C_SOURCE = "C_SOURCE"
    C_PREPROCESSED_SOURCE = "C_PREPROCESSED_SOURCE"
    CPP_SOURCE = "CPP_SOURCE"
    CPP_PREPROCESSED_SOURCE = "CPP_PREPROCESSED_SOURCE"

    JAVA_SOURCE = "JAVA_SOURCE"
    JAVA_BITCODE = "JAVA_BITCODE"

    PYTHON_SOURCE = "PYTHON_SOURCE"
    JAVASCRIPT_SOURCE = "JAVASCRIPT_SOURCE"

#~ class CodeFormats()

class BaseCodeFormatConverter(abc.ABC):
    @abc.abstractmethod
    def convert_code(self, src_fmt, dest_fmt, file_src_dest_map, \
                                                repository_manager, **kwargs):
        pass

    @abc.abstractmethod
    def get_source_formats(self):
        pass

    @abc.abstractmethod
    def get_destination_formats_for(self, src_fmt):
        pass
#~ class BaseCodeFormatConverter

class IdentityCodeConverter(BaseCodeFormatConverter):

    class CopyCallbackObject(DefaultCallbackObject):
        def after_command(self):
            file_src_dest_map = self.post_callback_args
            self._copy_from_repo(file_src_dest_map)
            return DefaultCallbackObject.after_command(self)
        #~ def after_command()
    #~ class CopyCallbackObject

    def convert_code(self, src_fmt, dest_fmt, file_src_dest_map, \
                                                repository_manager, **kwargs):
        # make sure formats are samel
        ERROR_HANDLER.assert_true(src_fmt == dest_fmt, \
                                                "different formats", __file__)
        # make sure that different sources have different destinations
        ERROR_HANDLER.assert_true(len(file_src_dest_map) == \
                    len({file_src_dest_map[fn] for fn in file_src_dest_map}), \
                        "Must specify one destination for each file", __file__)
        # copy the sources into the destinations
        copy_callback_obj = self.CopyCallbackObject()
        copy_callback_obj.set_post_callback_args(file_src_dest_map)

        b_ret, a_ret = repository_manager.custom_read_access(copy_callback_obj)
        ERROR_HANDLER.assert_true(\
                    b_ret == common_mix.GlobalConstants.COMMAND_SUCCESS & \
                    a_ret == common_mix.GlobalConstants.COMMAND_SUCCESS, \
                                                "code copy failed", __file__)
        return b_ret, common_mix.GlobalConstants.COMMAND_UNCERTAIN, a_ret
    #~ def identity_function()

    def get_source_formats(self):
        ERROR_HANDLER.error_exit(\
                    "get_source_formats must not be called here", __file__)
    #~ def get_source_formats()

    def get_destination_formats_for(self, src_fmt):
        ERROR_HANDLER.error_exit(\
                "get_destination_formats must not be called here", __file__)
    #~ def get_destination_formats()
#~ class IdentityCodeConverter



from __future__ import print_function

import os
import shutil
import logging
from jinja2 import Template
import webbrowser

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs
import muteria.common.matrices as common_matrices

import muteria.statistics.algorithms as stats_algo

import muteria.controller.explorer as fd_structure

ERROR_HANDLER = common_mix.ErrorHandler

def get_subsuming_elements(matrix_file):
    mat = common_matrices.ExecutionMatrix(filename=matrix_file)
    elem_to_tests = mat.query_active_columns_of_rows()
    equiv, subs_clusters = stats_algo.getSubsumingMutants(\
                                        elem_to_tests, clustered=True)
    return equiv, subs_clusters
#~ def get_subsuming_elements()

class StatsComputer(object):
    @staticmethod
    def merge_lmatrix_into_right(lmatrix_file, rmatrix_file):
        if not os.path.isfile(rmatrix_file):
            shutil.copy2(lmatrix_file, rmatrix_file)
        else:
            lmatrix = common_matrices.ExecutionMatrix(filename=lmatrix_file)
            rmatrix = common_matrices.ExecutionMatrix(filename=rmatrix_file)
            rmatrix.update_with_other_matrix(lmatrix, override_existing=True, \
                                    ask_confirmation_with_exist_missing=True)
            rmatrix.serialize()
    #~ def merge_lmatrix_into_right()

    @staticmethod
    def merge_lexecoutput_into_right(lexecoutput_file, rexecoutput_file):
        if not os.path.isfile(rexecoutput_file):
            shutil.copy2(lexecoutput_file, rexecoutput_file)
        else:
            lexecoutput = common_matrices.OutputLogData(\
                                                    filename=lexecoutput_file)
            rexecoutput = common_matrices.OutputLogData(\
                                                    filename=rexecoutput_file)
            rexecoutput.update_with_other(lexecoutput)
            rexecoutput.serialize()
    #~ def merge_lmatrix_into_right()

    @staticmethod
    def compute_stats(config, explorer, checkpointer):
        # get the matrix of each test criterion
        coverages = {}
        total_to = {}
        number_of_testcases = None
        for c in config.ENABLED_CRITERIA.get_val():
            if explorer.file_exists(fd_structure.CRITERIA_MATRIX[c]):
                mat_file = explorer.get_existing_file_pathname(\
                                            fd_structure.CRITERIA_MATRIX[c])
                mat = common_matrices.ExecutionMatrix(filename=mat_file)
                row2collist = mat.query_active_columns_of_rows()
                cov = len([k for k,v in row2collist.items() if len(v) > 0])
                tot = len(row2collist)
                coverages[c.get_str()] = 'n.a.' if tot == 0 else \
                                          '{:.2f}'.format(cov * 100.0 / tot)
                total_to[c.get_str()] = tot
                if number_of_testcases is None:
                    number_of_testcases = len(mat.get_nonkey_colname_list())
        
        # JSON
        out_json = {}
        out_json['TOTAL EXECUTION TIME (s)'] = \
                                            checkpointer.get_execution_time()
        out_json['NUMBER OF TESTCASES'] = number_of_testcases
        out_json['CRITERIA'] = {}
        for c in coverages:
            out_json['CRITERIA'][c] = {'coverage': coverages[c], 
                                            '# test objectives': total_to[c]}
        common_fs.dumpJSON(out_json, explorer.get_file_pathname(\
                                            fd_structure.STATS_MAIN_FILE_JSON))

        # HTML
        template_file = os.path.join(os.path.dirname(\
                            os.path.abspath(__file__)), 'summary_report.html')
        report_file = explorer.get_file_pathname(\
                                            fd_structure.STATS_MAIN_FILE_HTML)
        
        def format_execution_time(exec_time):
            n_day = int(exec_time // (24 * 3600))
            exec_time = exec_time % (24 * 3600)
            n_hour = int(exec_time // 3600)
            exec_time %= 3600
            n_minutes = int(exec_time // 60)
            exec_time %= 60
            n_seconds = int(round(exec_time))

            res = ""
            for val, unit in [(n_day, 'day'), (n_hour, 'hour'), \
                                (n_minutes, 'minutes'), (n_seconds, 'second')]:
                if val > 0:
                    s = ' ' if val == 1 else 's '
                    res += str(val) + ' ' + unit + s
            
            return res
        #~ def format_execution_time()

        total_exec_time = format_execution_time(\
                                            checkpointer.get_execution_time())

        rendered = Template(open(template_file).read()).render( \
                                {
                                    'total_execution_time': total_exec_time,
                                    'number_of_testcases': number_of_testcases,
                                    'coverages':coverages, 
                                    'total_to':total_to,
                                })
        with open(report_file, 'w') as f:
            f.write(rendered)
        
        try:
            webbrowser.get()
            webbrowser.open('file://' + report_file,new=2)
        except Exception as e:
            logging.warning("webbrowser error: "+str(e))
    #~ def compute_stats()
#~ class DataHandling




from __future__ import print_function
import os
import sys
import glob
import shutil
import logging
import re

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix
import muteria.common.matrices as common_matrices

import muteria.controller.explorer as fd_structure
import muteria.drivers.criteria as criteria

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.drivers.testgeneration.base_testcasetool import BaseTestcaseTool
from muteria.drivers.testgeneration.testcases_info import TestcasesInfoObject
from muteria.drivers import DriversUtils

from muteria.repositoryandcode.callback_object import DefaultCallbackObject

from muteria.drivers.testgeneration.tools_by_languages.c.klee.klee \
                                                    import TestcasesToolKlee

ERROR_HANDLER = common_mix.ErrorHandler

class TestcasesToolShadowSE(TestcasesToolKlee):
    """ Make sure to set the path to binarydir in user customs to use this
        The path to binary should be set to the path to the shadow 
        directory. in Shadow VM, it should be '/home/shadowvm/shadow'
    """
    def __init__(self, *args, **kwargs):
        TestcasesToolKlee.__init__(self, *args, **kwargs)
        ERROR_HANDLER.assert_true(self.custom_binary_dir is not None, \
                        "Custom binary dir must be set for shadow", __file__)
        self.shadow_folder_path = os.sep.join(\
                                    os.path.abspath(self.custom_binary_dir)\
                                    .split(os.sep)[:-3])
        self.llvm_29_compiler_path = os.path.join(self.shadow_folder_path, \
                                    "kleeDeploy/llvm-2.9/Release+Asserts/bin")
        self.llvm_gcc_path = os.path.join(self.shadow_folder_path, \
                                'kleeDeploy/llvm-gcc4.2-2.9-x86_64-linux/bin')
        self.wllvm_path = os.path.join(self.shadow_folder_path, \
                                            'kleeDeploy/whole-program-llvm')
        self.klee_change_locs_list_file = os.path.join(self.tests_working_dir,\
                                                    "klee_change_locs.json")
        # decide whether to replay directly the ktest 
        # or go through the test used during ktest generation
        self.replay_using_src_test = False

        # XXX: The first test generated by shadow is same as the seed used
        # to avoid duplication, set this to False so it is removed
        self.keep_first_test = False
    #~ def __init__()

    def set_keep_first_test(self, bool_val):
        ERROR_HANDLER.assert_true(type(bool_val) == bool, \
                            "invalid bool_val type. Must be bool", __file__)
        self.keep_first_test = bool_val
    #~ def set_keep_first_test()

    # SHADOW override
    def _get_default_params(self):
        bool_params = {
            '-ignore-solver-failures': None,
            '-allow-external-sym-calls': True, #None,
            '-posix-runtime': True, #None,
            '-dump-states-on-halt': True, #None,
            #'-only-output-states-covering-new': True 
            '--zest': True,
            '--shadow': True,
            '-emit-all-errors': True,
            '-no-std-out': True,
            '-shadow-allow-allocs': True,
            '-watchdog': True,
            '-shadow-replay-standalone': False,
            '-shadow-only-symbolic-tests': True,
        }
        key_val_params = {
            #'-output-dir': self.tests_storage_dir,
            #'-solver-backend': None,
            '-search': None,
            '-max-memory': None,
            '-max-time': self.config.TEST_GENERATION_MAXTIME,
            '-libc': 'uclibc',
            '-use-shadow-version': 'product',
            '-program-name': None, #TODO: CHeck importance
        }
        return bool_params, key_val_params
    #~ def _get_default_params()
    
    # SHADOW override
    def _get_sym_args(self):
        # sym args
        default_sym_args = [] #['-sym-arg', '5']
        klee_sym_args = default_sym_args

        #klee_sym_args = default_sym_args
        #uc = self.config.get_tool_user_custom()
        #if uc is not None:
        #    post_bc_cmd = uc.POST_TARGET_CMD_ORDERED_FLAGS_LIST
        #    if post_bc_cmd is not None:
        #        klee_sym_args = []
        #        for tup in post_bc_cmd:
        #            klee_sym_args += list(tup)
        return klee_sym_args
    #~ def _get_sym_args()

    # SHADOW should override
    def _get_back_llvm_compiler(self):
        # this is commented because the newest wllvm does not recognise 
        # llvm-gcc thus a symlink, named 'clang', to llvm-gcc 
        # is set in shadow folder
        #return "llvm-gcc"  
        return 'clang'
    #~ def _get_back_llvm_compiler()

    # SHADOW should override
    def _get_back_llvm_compiler_path(self):
        return self.llvm_29_compiler_path 
    #~ def _get_back_llvm_compiler_path()

    # SHADOW should override
    def _call_generation_run(self, runtool, args):
        # Delete any klee-out-*
        for d in os.listdir(self.tests_working_dir):
            if d.startswith('klee-out-'):
                shutil.rmtree(os.path.join(self.tests_working_dir, d))

        call_shadow_wrapper_file = os.path.join(self.tests_working_dir, \
                                                                "shadow_wrap")
        
        test_list = list(self.code_builds_factory.repository_manager\
                                                        .get_dev_tests_list())
        devtest_toolalias = self.parent_meta_tool.get_devtest_toolalias()

        # Get list of klee_change, klee_get_true/false locations.
        klee_change_stmts = []
        
        get_lines_callback_obj = self.GetLinesCallbackObject()
        get_lines_callback_obj.set_pre_callback_args(self.code_builds_factory\
                                    .repository_manager.revert_src_list_files)
        get_lines_callback_obj.set_post_callback_args(klee_change_stmts)

        pre_ret, post_ret = self.code_builds_factory.repository_manager\
                                    .custom_read_access(get_lines_callback_obj)
        ERROR_HANDLER.assert_true(pre_ret == \
                                common_mix.GlobalConstants.COMMAND_SUCCESS,\
                                                    "pre failed", __file__)
        ERROR_HANDLER.assert_true(post_ret == \
                                common_mix.GlobalConstants.COMMAND_SUCCESS,\
                                                    "post failed", __file__)

        ERROR_HANDLER.assert_true(len(klee_change_stmts) > 0, \
                        "No klee_change statement in the sources", __file__)

        # Filter only tests that cover those locations, 
        # if there is stmt coverage matrix 
        stmt_cov_mat_file = self.head_explorer.get_file_pathname(\
                            fd_structure.CRITERIA_MATRIX[criteria.TestCriteria\
                                                        .STATEMENT_COVERAGE])
        cov_tests = None
        if os.path.isfile(stmt_cov_mat_file):
            stmt_cov_mat = common_matrices.ExecutionMatrix(\
                                                    filename=stmt_cov_mat_file)
            # due to possible wrapper test splitting we update test_list here
            tmp_test_list = []
            for mt in stmt_cov_mat.get_nonkey_colname_list():
                alias, t = DriversUtils.reverse_meta_element(mt)
                if alias == devtest_toolalias:
                    tmp_test_list.append(t)
            test_list = tmp_test_list

            meta_stmts = list(stmt_cov_mat.get_keys())
            tool_aliases = set()
            for meta_stmt in meta_stmts:
                alias, stmt = DriversUtils.reverse_meta_element(meta_stmt)
                tool_aliases.add(alias)
            klee_change_meta_stmts = []
            for alias in tool_aliases:
                klee_change_meta_stmts += [\
                                    DriversUtils.make_meta_element(e, alias) \
                                                    for e in klee_change_stmts]
            klee_change_meta_stmts = list(set(meta_stmts) & \
                                                set(klee_change_meta_stmts))

            cov_tests = set()
            if len(klee_change_meta_stmts) > 0:
                for _, t in stmt_cov_mat.query_active_columns_of_rows(\
                                row_key_list=klee_change_meta_stmts).items():
                    cov_tests |= set(t)
            else:
                logging.warning('No test covers the patch (SHADOW)!')
            #    ERROR_HANDLER.assert_true(len(klee_change_meta_stmts) > 0, \
            #                            "No test covers the patch", __file__)

        # tests will be generated in the same dir that has the input .bc file
        os.mkdir(self.tests_storage_dir)

        # obtain candidate tests
        cand_testpair_list = []
        for test in test_list:
            meta_test = DriversUtils.make_meta_element(test, devtest_toolalias)
            if cov_tests is not None and meta_test not in cov_tests:
                continue 
            cand_testpair_list.append((test, meta_test))

        # Adjust the max-time in args
        ## locate max-time
        if len(cand_testpair_list) > 0:
            cur_max_time = float(self.get_value_in_arglist(args, 'max-time'))
            self.set_value_in_arglist(args, 'max-time', \
                        str(max(60, cur_max_time / len(cand_testpair_list))))

        # Set the wrapper
        with open(call_shadow_wrapper_file, 'w') as wf:
            wf.write('#! /bin/bash\n\n')
            wf.write('ulimit -s unlimited\n')
            wf.write(' '.join(['exec', runtool] + args + ['"${@:1}"']) + '\n')
        os.chmod(call_shadow_wrapper_file, 0o775)

        # run test
        exes, _ = self.code_builds_factory.repository_manager\
                                                .get_relative_exe_path_map()
        ERROR_HANDLER.assert_true(len(exes) == 1, \
                                            "Must have a single exe", __file__)
        exe_path_map = {e: call_shadow_wrapper_file for e in exes}
        env_vars = {}
        self._dir_chmod777(self.tests_storage_dir)
        for test, meta_test in cand_testpair_list:
            self.parent_meta_tool.execute_testcase(meta_test, exe_path_map, \
                                        env_vars, with_output_summary=False)

            # copy the klee out
            test_out = os.path.join(self.tests_storage_dir, \
                                                    test.replace(os.sep, '_'))
            os.mkdir(test_out)
            for d in glob.glob(self.tests_working_dir+"/klee-out-*"):
                # make sure we can do anything with it
                self._dir_chmod777(d)
                if not self.keep_first_test:
                    first_test = os.path.join(d, 'test000001.ktest')
                    if os.path.isfile(first_test):
                        shutil.move(first_test, first_test+'.disable')
                shutil.move(d, test_out)
            ERROR_HANDLER.assert_true(len(list(os.listdir(test_out))) > 0, \
                                "Shadow generated no test for tescase: "+test,\
                                                                    __file__)
            if os.path.islink(os.path.join(self.tests_working_dir, \
                                                                'klee-last')):
                os.unlink(os.path.join(self.tests_working_dir, 'klee-last'))

        # store klee_change locs
        common_fs.dumpJSON(klee_change_stmts, self.klee_change_locs_list_file)
    #~ def _call_generation_run()

    class GetLinesCallbackObject(DefaultCallbackObject):
        def before_command(self):
            revert_src_func = self.pre_callback_args
            revert_src_func()
            return common_mix.GlobalConstants.COMMAND_SUCCESS
        #~ def before_command()

        def after_command(self):
            if self.op_retval == common_mix.GlobalConstants.COMMAND_FAILURE:
                return common_mix.GlobalConstants.COMMAND_FAILURE
            m_regex = re.compile('({}|{}|{})'.format('klee_change', \
                                            'klee_get_true', 'klee_get_false'))
            matched_lines = set()
            for src in self.source_files_to_objects:
                with open(os.path.join(self.repository_rootdir, src), \
                                                        encoding='UTF-8') as f:
                    for lnum, line in enumerate(f.readlines()):
                        if m_regex.search(line) is not None:
                            matched_lines.add(DriversUtils.make_meta_element(\
                                                            str(lnum+1), src))
            ret_lines = self.post_callback_args
            ret_lines.extend(matched_lines)
            return common_mix.GlobalConstants.COMMAND_SUCCESS
        #~ def after_command()
    #~ class CopyCallbackObject

    # SHADOW should override
    def _get_testexec_extra_env_vars(self, testcase):
        return None # TODO: get test env and return
    #~ def _get_testexec_extra_env_vars()

    def _do_generate_tests (self, exe_path_map, code_builds_factory, \
                                                meta_criteria_tool_obj=None, \
                                                                max_time=None):
        env_path_bak = os.environ['PATH']
        os.environ['PATH'] = os.pathsep.join([self.llvm_gcc_path, \
                                         self.wllvm_path, os.environ['PATH']])

        super(TestcasesToolShadowSE, self)._do_generate_tests(\
                                        exe_path_map, code_builds_factory, \
                            meta_criteria_tool_obj=meta_criteria_tool_obj, \
                                                            max_time=max_time)

        os.environ['PATH'] = env_path_bak
    #~ def _do_generate_tests ()
#~ class TestcasesToolShadowSE



""" Put in here the extra configuration that can be passed for a 
specific part of the tool. 
Example: linking flags for compiling llvm module into executable

The format is <Component using>__<Module of Functionality>__<Parameter name>
"""

CODE_BUILD_FACTORY__LLVM__LINKING_FLAGS = None



from __future__ import print_function

import os
import sys
import shutil
import logging
import abc

import muteria.common.mix as common_mix

import muteria.drivers.testgeneration.custom_dev_testcase.system_wrappers as \
                                                                system_wrappers

ERROR_HANDLER = common_mix.ErrorHandler

class BaseSystemTestSplittingWrapper(abc.ABC):
    def get_sub_test_id_env_vars(self, subtest_id):
        return {system_wrappers.TEST_COUNT_ID_ENV_VAR: str(subtest_id)}    
    #~ def get_sub_test_id_env_vars()

    @abc.abstractmethod
    def set_wrapper(self, workdir, exe_path_map):
        """ Return the new exe path map
        """
        print ("Implement!!!")
    #~ def set_wrapper()

    @abc.abstractmethod
    def switch_to_new_test(self):
        """ reset the counters
        """
        print ("Implement!!!")
    #~ def switch_to_new_test()

    @abc.abstractmethod
    def collect_data(self):
        """ get number of sub tests and args
        """
        print ("Implement!!!")
    #~ def collect_data()

    @abc.abstractmethod
    def cleanup(self):
        print ("Implement!!!")
    #~ def cleanup()
#~ class BaseSystemTestSplittingWrapper

class BaseSystemWrapper(abc.ABC):
    
    # do not change

    backup_ext = '.muteria_bak'
    used_ext = '.muteria_used'

    counter_ext = '.muteria_counter'

    outlog_ext = '.muteria_outlog'
    outretcode_ext = '.muteria_outretcode'

    # Must Override

    @abc.abstractmethod
    def get_dev_null(self):
        print ("Implement!!!")
    #~ def get_dev_null()

    @abc.abstractmethod
    def _get_wrapper_template_string(self):
        print("Implement!!!")
    #~ def _get_wrapper_template_string()

    @abc.abstractmethod
    def _get_timedout_codes(self):
        print("Implement!!!")
    #~ def _get_timedout_codes()

    ## Wrapper test splitting TODO TODO
    @abc.abstractmethod
    def get_test_splitting_wrapper_class(self):
        print("Implement!!!")
    #~ def get_sub_test_id_env_vars()

    # Can override

    def __init__(self, repo_mgr):
        self.repo_mgr = repo_mgr
        self.test_splitting_wrapper = self.get_test_splitting_wrapper_class()
        if self.test_splitting_wrapper is not None:
            self.test_splitting_wrapper = self.test_splitting_wrapper()
    #~ def __init__()

    def get_test_splitting_wrapper(self):
        return self.test_splitting_wrapper
    #~ def get_test_splitting_wrapper()

    def _get_repo_run_path_pairs(self, exe_path_map):
        ERROR_HANDLER.assert_true(len(exe_path_map) == 1, \
                    "support a single exe for now. got: "+str(exe_path_map), \
                                                                    __file__)
        repo_exe = list(exe_path_map.keys())[0]
        run_exe = exe_path_map[repo_exe]
        repo_exe = self.repo_mgr.repo_abs_path(repo_exe)
        if run_exe is None:
            run_exe = repo_exe
        return [(repo_exe, run_exe)]
    #~ def _get_repo_run_path_pairs()

    def cleanup_logs(self, exe_path_map):
        repo_exe_abs_path, _ = self._get_repo_run_path_pairs(exe_path_map)[0]
        if os.path.isfile(repo_exe_abs_path + self.counter_ext):
            os.remove(repo_exe_abs_path + self.counter_ext)
        if os.path.isfile(repo_exe_abs_path + self.outretcode_ext):
            os.remove(repo_exe_abs_path + self.outretcode_ext)
        if os.path.isfile(repo_exe_abs_path + self.outlog_ext):
            os.remove(repo_exe_abs_path + self.outlog_ext)
    #~ def cleanup(repo_exe_abs_path):

    def collect_output(self, exe_path_map, collected_output, testcase):
        repo_exe_abs_path, _ = self._get_repo_run_path_pairs(exe_path_map)[0]
        tmp = []
        if not os.path.isfile(repo_exe_abs_path + self.outretcode_ext):
            ERROR_HANDLER.error_exit("testcase has no log: '" +testcase+ "'."
                                " repo_exe_abs_path is " + repo_exe_abs_path, \
                                                                    __file__)
        timedout = []

        with open(repo_exe_abs_path + self.outretcode_ext) as f:
            for line in f:
                tmp.append(int(line.strip()))
                timedout.append(tmp[-1] in self._get_timedout_codes())            
            if len(tmp) == 1:
                tmp = tmp[0]
                timedout = timedout[0]
        collected_output.append(tmp)

        try:
            with open(repo_exe_abs_path + self.outlog_ext) as f:
                collected_output.append(f.read())
        except UnicodeDecodeError:
            with open(repo_exe_abs_path + self.outlog_ext, \
                                                encoding='ISO-8859-1') as f:
                collected_output.append(f.read())

        collected_output.append(timedout) 
    #~ def collect_output()

    def install_wrapper(self, exe_path_map, collect_output):
        repo_exe_abs_path, run_exe_abs_path = \
                                self._get_repo_run_path_pairs(exe_path_map)[0]

        if os.path.isfile(repo_exe_abs_path + self.used_ext):
            os.remove(repo_exe_abs_path + self.used_ext)

        # set run exe
        try:
            shutil.copy2(run_exe_abs_path, repo_exe_abs_path + self.used_ext)
            # use link instead of copy to avoid copying large unchanging exes
            #os.link(run_exe_abs_path, repo_exe_abs_path + self.used_ext)
        except PermissionError:
            os.remove(repo_exe_abs_path + self.used_ext)
            shutil.copy2(run_exe_abs_path, repo_exe_abs_path + self.used_ext)
            # use link instead of copy to avoid copying large unchanging exes
            #os.link(run_exe_abs_path, repo_exe_abs_path + self.used_ext)

        # backup
        shutil.move(repo_exe_abs_path, repo_exe_abs_path + self.backup_ext)

        # place the wrapper
        match_replacing = {
            'WRAPPER_TEMPLATE_DEFAULT_EXE_ASBSOLUTE_PATH': \
                                            repo_exe_abs_path+self.backup_ext,
            'WRAPPER_TEMPLATE_RUN_EXE_ASBSOLUTE_PATH': \
                                            repo_exe_abs_path+self.used_ext,
            'WRAPPER_TEMPLATE_COUNTER_FILE': \
                                        repo_exe_abs_path+self.counter_ext, \
            'WRAPPER_TEMPLATE_OUTPUT_RETCODE': \
                                    repo_exe_abs_path+self.outretcode_ext \
                                    if collect_output else self.get_dev_null(),
            'WRAPPER_TEMPLATE_OUTPUT_LOG': repo_exe_abs_path+self.outlog_ext \
                                    if collect_output else self.get_dev_null(),
        }
        wrapper_obj = self._get_wrapper_template_string()
        for match, replace in match_replacing.items():
            wrapper_obj = wrapper_obj.replace(match, replace)
        with open(repo_exe_abs_path, 'w') as dest:
            dest.write(wrapper_obj+'\n')

        # make executable
        shutil.copymode(repo_exe_abs_path + self.backup_ext, repo_exe_abs_path)

        # cleanup data
        self.cleanup_logs(exe_path_map)
    #~ def install_wrapper()

    def uninstall_wrapper(self, exe_path_map):
        repo_exe_abs_path, _ = self._get_repo_run_path_pairs(exe_path_map)[0]

        # unset run exe
        shutil.move(repo_exe_abs_path + self.backup_ext, repo_exe_abs_path)

        # small cleanup
        if os.path.isfile(repo_exe_abs_path + self.used_ext):
            os.remove(repo_exe_abs_path + self.used_ext)

        # cleanup data
        self.cleanup_logs(exe_path_map)
    #~ def uninstall_wrapper()
#~ class BaseSystemWrapper




from __future__ import print_function

import os
import sys
import re
import shutil
import glob
import logging
import configparser

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.repositoryandcode.callback_object import DefaultCallbackObject

from muteria.drivers.criteria.base_testcriteriatool import BaseCriteriaTool
from muteria.drivers.criteria import TestCriteria
from muteria.drivers import DriversUtils

try:
    import coverage
except ImportError:
    pass

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaToolCoveragePy(BaseCriteriaTool):
    def __init__(self, *args, **kwargs):
        BaseCriteriaTool.__init__(self, *args, **kwargs)
        self.instrumentation_details = os.path.join(\
                    self.instrumented_code_storage_dir, '.instru.meta.json')
        self.cov_data_filename = "cov_data.json"
        self.used_srcs_dir = os.path.join(self.instrumented_code_storage_dir,\
                                                            'used_srcs_dir')
        self.preload_dir = os.path.join(self.instrumented_code_storage_dir,\
                                                                'config_dir')
        self.preload_file = os.path.join(self.preload_dir, 'usercustomize.py')
        self.config_file = os.path.join(self.instrumented_code_storage_dir,\
                                                                '.configrc')
        self.raw_data_file = os.path.join(self.instrumented_code_storage_dir,\
                                                                    '.rawdata')
        # clean any possible raw data file
        for file_ in glob.glob(self.raw_data_file+"*"):
            os.remove(file_)
    #~ def __init__()

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        try:
            import coverage
        except ImportError:
            return False
        #for prog in ('coverage'):
        #    if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
        #                                            expected_exit_codes=[0]):
        #        return False
        return True
    #~ def installed()

    @classmethod
    def _get_meta_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in same file
            :return: list of citeria
        """
        return [
                TestCriteria.STATEMENT_COVERAGE,
                TestCriteria.BRANCH_COVERAGE,
               ]
    #~ def _get_meta_instrumentation_criteria()

    @classmethod
    def _get_separated_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in different files
            :return: list of citeria
        """
        return []
    #~ def _get_separated_instrumentation_criteria()

    def get_instrumented_executable_paths_map(self, enabled_criteria):
        crit_to_exes_map = {}
        obj = common_fs.loadJSON(self.instrumentation_details)
        #exes = [p for _, p in list(obj.items())]
        exes = obj
        for criterion in enabled_criteria:
            crit_to_exes_map[criterion] = exes
        return crit_to_exes_map
    #~ def get_instrumented_executable_paths_map()

    def get_criterion_info_object(self, criterion):
        return None
    #~ def get_criterion_info_object(self, criterion)
    
    def _get_criterion_element_executable_path(self, criterion, element_id):
        ERROR_HANDLER.error_exit("not applicable for coverage_py", __file__)
    #~ def _get_criterion_element_executable_path

    def _get_criterion_element_environment_vars(self, criterion, element_id):
        '''
            return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        ERROR_HANDLER.error_exit("not applicable for coverage_py", __file__)
    #~ def _get_criterion_element_environment_vars()

    def _get_criteria_environment_vars(self, result_dir_tmp, enabled_criteria):
        '''
        return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        python_path = self.preload_dir
        if 'PYTHONPATH' is os.environ:
            python_path += ":"+os.environ['PYTHONPATH']

        return {criterion: {
                    #"PYTHONUSERBASE": self.preload_dir,
                    "PYTHONPATH": python_path, 
                    "COVERAGE_PROCESS_START": self.config_file,
                } for criterion in enabled_criteria}
    #~ def _get_criteria_environment_vars()

    class PathAliases(object):
        def __init__(self, data_files, exe_rel_files, inst_top_dir, \
                                                        top_out_dir, repo_dir):
            repo_dir = os.path.join(os.path.normpath(repo_dir), '')
            top_out_dir = os.path.join(os.path.normpath(top_out_dir), '')
            prefix_dirs = (repo_dir, top_out_dir)
            self.alias_map = {}
            #a_data_file = os.path.normpath(data_files[0])
            #prefix_out = prefix_repo = None
            #for fn in exe_rel_files:
            #    if a_data_file.endswith(fn):
            #        prefix = a_data_file[:-len(fn)]
            #ERROR_HANDLER.assert_true(prefix is not None, \
            #                "file in data has no match in source", __file__)
            for _, dfile in enumerate(data_files):
                src = tmp =None
                for p_dir in prefix_dirs:
                    if dfile.startswith(p_dir):
                        tmp = dfile[len(p_dir):]
                        if tmp in exe_rel_files:
                            src = tmp
                            break
                if src is None:
                    ERROR_HANDLER.assert_true(tmp is None, \
                                    'The file {} {} {} {}'.format(\
                                    dfile, 'is in the right dirs and in',\
                                    'coverage but was not specified.',\
                                    'Bug in coverage.py?'), __file__)
                    self.alias_map[dfile] = dfile
                else:
                    self.alias_map[dfile] = os.path.join(inst_top_dir, src)

        def map(self, in_dat_file):
            return self.alias_map[in_dat_file]
    #~ PathAliases

    def _collect_temporary_coverage_data(self, criteria_name_list, \
                                                test_execution_verdict, \
                                                used_environment_vars, \
                                                result_dir_tmp, \
                                                testcase):
        ''' extract coverage data into json file in result_dir_tmp
        '''
        cov_obj = coverage.Coverage(config_file=self.config_file)
        cov_obj.combine()
        tmp_dat_obj = cov_obj.get_data()
        
        in_dat_files = tmp_dat_obj.measured_files()

        # Get file map
        try :
            self.exes_rel
        except AttributeError:
            obj = common_fs.loadJSON(self.instrumentation_details)
            self.exes_abs = []
            self.exes_rel = []
            for rp, ap in list(obj.items()):
                self.exes_rel.append(rp)
                self.exes_abs.append(ap)
            self.exes_rel, self.exes_abs = zip(*sorted(\
                                            zip(self.exes_rel, self.exes_abs),\
                                            key=lambda x: x.count(os.path.sep)\
                                            ))
            
            # get executables stmt and branches
            self.executable_lines = {}
            self.executable_arcs = {}
            for fn in self.exes_abs:
                pser = coverage.parser.PythonParser(filename=fn)
                pser.parse_source() 
                self.executable_lines[fn] = pser.statements
                self.executable_arcs[fn] = pser.arcs() 


        dat_obj = coverage.CoverageData()
        if len(in_dat_files) > 0:
            file_map = self.PathAliases(in_dat_files, self.exes_rel, \
                                    self.used_srcs_dir, \
                                    self._get_latest_top_output_dir(), \
                                    self.code_builds_factory.\
                                        repository_manager.repo_abs_path(''))
            dat_obj.update(tmp_dat_obj, aliases=file_map)

        # Get the coverages
        res = {c: {} for c in criteria_name_list}
        if TestCriteria.STATEMENT_COVERAGE in criteria_name_list:
            for rel_fn, abs_fn in zip(self.exes_rel, self.exes_abs):
                res[TestCriteria.STATEMENT_COVERAGE][rel_fn] = \
                                {ln: 0 for ln in self.executable_lines[abs_fn]}
                ln_list = dat_obj.lines(abs_fn)
                if ln_list is not None:
                    res[TestCriteria.STATEMENT_COVERAGE][rel_fn].\
                                            update({ln:1 for ln in ln_list})
        if TestCriteria.BRANCH_COVERAGE in criteria_name_list:
            for rel_fn, abs_fn in zip(self.exes_rel, self.exes_abs):
                res[TestCriteria.BRANCH_COVERAGE][rel_fn] = \
                            {str(an): 0 for an in self.executable_arcs[abs_fn]}
                an_list = dat_obj.arcs(abs_fn)
                if an_list is not None:
                    res[TestCriteria.BRANCH_COVERAGE][rel_fn].\
                                        update({str(an):1 for an in an_list})

        # save the temporary coverage
        res_str = {v.get_str():res[v] for v in res}
        common_fs.dumpJSON(res_str, os.path.join(result_dir_tmp,\
                                                    self.cov_data_filename))

        cov_obj.erase()
        # clean any possible raw data file
        for file_ in glob.glob(self.raw_data_file+"*"):
            os.remove(file_)
    #~ def _collect_temporary_coverage_data()

    def _extract_coverage_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        ''' read json files and extract data
            return: the dict of criteria with covering count
        '''
        in_file = os.path.join(result_dir_tmp, self.cov_data_filename)
        cov_dat_obj = common_fs.loadJSON(in_file)
        cov_dat_obj = {TestCriteria[v]: cov_dat_obj[v] for v in cov_dat_obj}

        ERROR_HANDLER.assert_true(set(cov_dat_obj) == set(enabled_criteria), \
                                    "mismatching criteria enabled", __file__)

        res = {c: {} for c in enabled_criteria}

        for c in cov_dat_obj:
            for filename, filedat in list(cov_dat_obj[c].items()):
                if filedat is not None:
                    for id_, cov in list(filedat.items()):
                        ident = DriversUtils.make_meta_element(\
                                                            str(id_), filename)
                        res[c][ident] = cov

        # delete cov file
        os.remove(in_file)

        return res
    #~ def _extract_coverage_data_of_a_test()

    def _do_instrument_code (self, exe_path_map, \
                                        code_builds_factory, \
                                        enabled_criteria, parallel_count=1):
        # Setup
        if os.path.isdir(self.instrumented_code_storage_dir):
            shutil.rmtree(self.instrumented_code_storage_dir)
        os.mkdir(self.instrumented_code_storage_dir)
        if os.path.isdir(self.used_srcs_dir):
            shutil.rmtree(self.used_srcs_dir)
        os.mkdir(self.used_srcs_dir)
        if os.path.isdir(self.preload_dir):
            shutil.rmtree(self.preload_dir)
        os.mkdir(self.preload_dir)

        rel_path_map = {}
        exes, src_inter_map = code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
        for f in src_inter_map:
            rel_path_map[f] = os.path.join(self.used_srcs_dir, f)
            relloc = os.path.join(self.used_srcs_dir, os.path.dirname(f))
            if not os.path.isdir(relloc):
                os.makedirs(relloc)
        not_none_dest = {k:v for k,v in list(rel_path_map.items())}
        pre_ret, ret, post_ret = code_builds_factory.transform_src_into_dest(\
                        src_fmt=CodeFormats.PYTHON_SOURCE,\
                        dest_fmt=CodeFormats.PYTHON_SOURCE,\
                        src_dest_files_paths_map=not_none_dest)
        # write down the rel_path_map
        ERROR_HANDLER.assert_true(not os.path.isfile(\
                self.instrumentation_details), "must not exist here", __file__)
        common_fs.dumpJSON(rel_path_map, self.instrumentation_details)

        if ret == common_mix.GlobalConstants.COMMAND_FAILURE:
            ERROR_HANDLER.error_exit("Problem with copying python sources", \
                                                                    __file__)

        # Create config and preload
        with open(self.preload_file, "w") as f:
            f.write("import coverage\ncoverage.process_startup()\n")
        
        concurrencies = ['thread', 'multiprocessing', 'gevent', 'greenlet', \
                                                                'eventlet']
        concurrency_support = 'thread'

        #config = configparser.ConfigParser()
        config = {}
        config['run'] = {
                        'include': \
                            [os.path.join('*',v) for v in rel_path_map.keys()],
                        'data_file': self.raw_data_file,
                        'branch': (\
                            TestCriteria.BRANCH_COVERAGE in enabled_criteria),
                        'parallel': True,
                        'concurrency': concurrency_support,
                        }
        with open(self.config_file, 'w') as f:
            #config.write(f)
            for k, v in list(config.items()):
                f.write('['+k+']'+'\n')
                for kk, vv in list(v.items()):
                    if type(vv) == list:
                        f.write(kk+' = \n')
                        for vvv in vv:
                            f.write('\t'+str(vvv)+'\n')
                    else:
                        f.write(kk+' = '+str(vv)+'\n')
    #~ def _do_instrument_code()
#~ class CriteriaToolCoveragePy



import logging

import muteria.common.mix as common_mix

ERROR_HANDLER = common_mix.ErrorHandler

class CheckPointHandler(object):
    FUNC_NAME_KEY = "method_name"
    TASK_ID_KEY = "task_id"
    TOOLS_KEY = "tools"
    OPT_PAYLOAD_KEY = "optional_payload"
    def __init__(self, used_checkpointer):
        self.used_checkpointer = used_checkpointer
        self.current_data = None
        self.get_current_data()

    def get_current_data(self):
        if self.current_data is None:
            self.current_data = \
                        self.used_checkpointer.load_checkpoint_or_start()
            # verify
            if self.current_data is not None:
                if type(self.current_data) != dict:
                    ERROR_HANDLER.error_exit( \
                        "Invalid checkpoint data. Must be a dict", __file__)
                if set(self.current_data) != {self.FUNC_NAME_KEY, \
                                        self.TASK_ID_KEY, self.TOOLS_KEY, \
                                        self.OPT_PAYLOAD_KEY}:
                    ERROR_HANDLER.error_exit( \
                       "Problem with checkpoint data dict key", __file__)

        return self.current_data

    def is_finished(self):
        return self.used_checkpointer.is_finished()

    def is_to_execute(self, func_name, taskid, tool=None):
        dat = self.get_current_data()
        if dat is None:
            return True
        if func_name != self.current_data[self.FUNC_NAME_KEY]:
            return True
        if taskid >= self.current_data[self.TASK_ID_KEY]:
            if taskid > self.current_data[self.TASK_ID_KEY] or \
                             tool not in self.current_data[self.TOOLS_KEY]:
                return True
        return False

    def do_checkpoint(self, func_name, taskid, tool=None, opt_payload=None):
        if self.current_data is None:
            self.current_data = { \
                self.FUNC_NAME_KEY: None, \
                self.TASK_ID_KEY: -1, \
                self.TOOLS_KEY: [], \
                self.OPT_PAYLOAD_KEY: None
            }
        self.current_data[self.FUNC_NAME_KEY] = func_name
        if self.current_data[self.TASK_ID_KEY] != taskid:
            self.current_data[self.TOOLS_KEY] = []
        self.current_data[self.TASK_ID_KEY] = taskid
        self.current_data[self.TOOLS_KEY].append(tool)
        self.current_data[self.OPT_PAYLOAD_KEY] = opt_payload
        self.used_checkpointer.write_checkpoint(self.current_data)

    def set_finished(self, detailed_exectime_obj):
        self.used_checkpointer.set_finished( \
                                detailed_exectime_obj=detailed_exectime_obj)

    def get_optional_payload(self):
        if self.current_data is None:
            return None
        return self.current_data[self.OPT_PAYLOAD_KEY]

    def restart(self):
        self.used_checkpointer.restart_task()

    def destroy(self):
        self.used_checkpointer.destroy_checkpoint()
#~ class CheckPointHandler





from muteria.drivers.criteria.tools_by_languages.python.coverage_py.coverage \
                                                import CriteriaToolCoveragePy 

#from .gcov import CriteriaToolGCov

StaticCriteriaTool = CriteriaToolCoveragePy




from __future__ import print_function

import importlib

import muteria.common.mix as common_mix

import muteria.drivers.optimizers.testexecution.tools as test_opt

class TestOptimizers(common_mix.EnumAutoName):
    # Default (No Optimization)
    NO_OPTIMIZATION = importlib.import_module(".default", \
                                                package=test_opt.__name__
                                            ).TestExecutionOptimizer

    def get_optimizer(self):
        return self.get_field_value()
    #~ def get_optimizer():
#~ class TestOptimizers



from muteria.drivers.testgeneration.tools_by_languages.c.klee.klee \
                                                    import TestcasesToolKlee 
StaticTestcaseTool = TestcasesToolKlee
DynamicTestcaseTool = None


"""
"""

from __future__ import print_function

import sys
import os
import logging
import shutil

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

import muteria.repositoryandcode.code_transformations as ct_modules
import muteria.repositoryandcode.codes_convert_support as ccs

from muteria.repositoryandcode.callback_object import DefaultCallbackObject

ERROR_HANDLER = common_mix.ErrorHandler

formatfrom_function_tuples = [
    # From C and CPP source
    (ccs.CodeFormats.C_SOURCE, ct_modules.c_cpp.FromC()),
    (ccs.CodeFormats.C_PREPROCESSED_SOURCE, ct_modules.c_cpp.FromC()),
    (ccs.CodeFormats.CPP_SOURCE, ct_modules.c_cpp.FromCpp()),
    (ccs.CodeFormats.CPP_PREPROCESSED_SOURCE, ct_modules.c_cpp.FromCpp()),

    # From LLVM bitcode
    (ccs.CodeFormats.LLVM_BITCODE, ct_modules.llvm.FromLLVMBitcode()),

    # From Javascript source
    (ccs.CodeFormats.JAVASCRIPT_SOURCE, ccs.IdentityCodeConverter()),

    # From Python source
    (ccs.CodeFormats.PYTHON_SOURCE, ccs.IdentityCodeConverter()),
]

class CodeBuildsFactory(object):
    def __init__(self, repository_manager, workdir=None):
        self.repository_manager = repository_manager
        self.workdir = workdir
        self.src_dest_fmt_to_handling_obj = {}

        self.code_conversion_tracker_file = None
        self.stored_files_mapping = None
        self.stored_files_mapping_file = None
        if self.workdir is not None:
            self.repo_path_saved_file = os.path.join(self.workdir, \
                                                             "saved_repo_path")
            self.code_conversion_tracker_file = \
                                    os.path.join(self.workdir, "is_converting")
            self.stored_files_mapping_file = \
                                        os.path.join(self.workdir, "files_map")
            exes, src_obj_map = \
                            self.repository_manager.get_relative_exe_path_map()
            if os.path.isdir(self.workdir):
                tmp = common_fs.loadJSON(self.stored_files_mapping_file)
                self.stored_files_mapping = {k: os.path.join(self.workdir, v) \
                                                        for k,v in tmp.items()}
            else:
                count = 0
                self.stored_files_mapping = {}
                tmp = {}
                for f in exes:
                    tmp[f] = str(count)
                    self.stored_files_mapping[f] = \
                                        os.path.join(self.workdir, str(count))
                    count += 1
                for s,o in list(src_obj_map.items()):
                    tmp[o] = str(count)
                    self.stored_files_mapping[o] = \
                                        os.path.join(self.workdir, str(count))
                    count += 1
                os.mkdir(self.workdir)
                common_fs.dumpJSON(tmp, \
                                self.stored_files_mapping_file, pretty=True)
        
        # Initialize 
        for src_fmt, obj_cls in formatfrom_function_tuples:
            if isinstance(obj_cls, ccs.IdentityCodeConverter):
                self._fmt_from_to_registration(src_fmt, src_fmt, obj_cls)
            else:
                ERROR_HANDLER.assert_true(\
                                src_fmt in obj_cls.get_source_formats(), \
                                "{} {} {} {}".format( \
                            "Error in 'formatfrom_function_tuples'",
                            "src_fmt", src_fmt, "not in corresponding obj..."),
                                                                    __file__)
                for dest_fmt in obj_cls.get_destination_formats_for(src_fmt):
                    self._fmt_from_to_registration(src_fmt, dest_fmt, obj_cls)


    #~ def __init__()

    def _fmt_from_to_registration(self, src_fmt, dest_fmt, handling_obj):
        if src_fmt not in self.src_dest_fmt_to_handling_obj:
            self.src_dest_fmt_to_handling_obj[src_fmt] = {}
        ERROR_HANDLER.assert_true( \
                dest_fmt not in self.src_dest_fmt_to_handling_obj[src_fmt],
                "dest_fmt {} added twice for same src_fmt {}".format( \
                                                src_fmt, dest_fmt), __file__)
        self.src_dest_fmt_to_handling_obj[src_fmt][dest_fmt] = handling_obj
    #~ def _fmt_from_to_registration()

    def transform_src_into_dest (self, src_fmt, dest_fmt, \
                                        src_dest_files_paths_map, **kwargs):
        # Checks
        ERROR_HANDLER.assert_true( \
                            src_fmt in self.src_dest_fmt_to_handling_obj, \
                            "src_fmt {} not supported yet.".format(src_fmt), \
                                                                    __file__)
        ERROR_HANDLER.assert_true( \
                    dest_fmt in self.src_dest_fmt_to_handling_obj[src_fmt], \
                    "dest_fmt {} not supported yet for src_fmt {}.".format( \
                                                dest_fmt, src_fmt), __file__)
        
        # call handler
        handler = self.src_dest_fmt_to_handling_obj[src_fmt][dest_fmt]

        # track code conversion
        if self.repository_manager.should_build():
            with open(self.code_conversion_tracker_file, 'w') as f:
                f.write('converting code...')

        pre_ret, ret, post_ret = handler.convert_code(src_fmt, dest_fmt, \
                            src_dest_files_paths_map, \
                            repository_manager=self.repository_manager, \
                            **kwargs)

        # untrack code conversion
        if self.repository_manager.should_build():
            os.remove(self.code_conversion_tracker_file)

        return pre_ret, ret, post_ret
    #~ def transform_src_into_dest ()
    
    def override_registration (self, src_fmt, dest_fmt, handling_obj):
        """set another obj to handle src dest pair or a new one
        """
        # invalidate existing
        if src_fmt in self.src_dest_fmt_to_handling_obj:
            if dest_fmt in self.src_dest_fmt_to_handling_obj[src_fmt]:
                del self.src_dest_fmt_to_handling_obj[src_fmt][dest_fmt]

        # register
        self._fmt_from_to_registration(src_fmt, dest_fmt, handling_obj)
    #~ def override_registration ()

    ## Useful
    
    class CopyCallbackObject(DefaultCallbackObject):
        def before_command(self):
            revert_src_func = self.pre_callback_args
            revert_src_func()
            return common_mix.GlobalConstants.COMMAND_SUCCESS
        #~ def before_command()

        def after_command(self):
            if self.op_retval == common_mix.GlobalConstants.COMMAND_FAILURE:
                return common_mix.GlobalConstants.COMMAND_FAILURE
            file_src_dest_map, reverse = self.post_callback_args
            if reverse:
                self._copy_to_repo(file_src_dest_map[0])
                # Do not consider the second
            else:
                self._copy_from_repo(file_src_dest_map[0])
                self._copy_from_repo(file_src_dest_map[1])
            return common_mix.GlobalConstants.COMMAND_SUCCESS
        #~ def after_command()
    #~ class CopyCallbackObject

    def set_repo_to_build_default(self, also_copy_to_map={}):
        if self.repository_manager.should_build():
            files_backed = False
            if self.stored_files_mapping is not None and \
                                            len(self.stored_files_mapping) > 0:
                # rebuild if failed while building
                if os.path.isfile(self.code_conversion_tracker_file):
                    for _, savef in self.stored_files_mapping.items():
                        if os.path.isfile(savef):
                            os.remove(savef)

                files_backed = os.path.isfile(\
                                    self.stored_files_mapping[\
                                    list(self.stored_files_mapping.keys())[0]])

                # copy the sources into the destinations
                copy_callback_obj = self.CopyCallbackObject()
                copy_callback_obj.set_pre_callback_args( \
                                self.repository_manager.revert_src_list_files)

            new_repo_path = self.repository_manager.get_repository_dir_path()
            if os.path.isfile(self.repo_path_saved_file):
                with open(self.repo_path_saved_file) as f:
                    old_repo_path = f.read().strip()
                if new_repo_path != old_repo_path: 
                    # Override potential backup (force rebuild)
                    files_backed = False
            else:
                # Override potential backup (force rebuild)
                files_backed = False

            if files_backed:
                copy_callback_obj.set_post_callback_args([\
                                (self.stored_files_mapping, also_copy_to_map),\
                                                                        True])
                b_ret, a_ret = self.repository_manager.custom_read_access(\
                                                            copy_callback_obj)
                ERROR_HANDLER.assert_true(\
                        b_ret == common_mix.GlobalConstants.COMMAND_SUCCESS & \
                        a_ret == common_mix.GlobalConstants.COMMAND_SUCCESS, \
                                                "code copy failed", __file__)
                # copy also
                for src,dest in also_copy_to_map.items():
                    try:
                        shutil.copy2(self.stored_files_mapping[src], dest)
                    except PermissionError:
                        os.remove(dest)
                        shutil.copy2(self.stored_files_mapping[src], dest)
            else:
                # build and possibly backup
                if self.stored_files_mapping is None:
                    #co = None
                    copy_callback_obj.set_post_callback_args([\
                                               ({}, also_copy_to_map), False])
                    co = copy_callback_obj
                else:
                    copy_callback_obj.set_post_callback_args([\
                                (self.stored_files_mapping, also_copy_to_map),\
                                                                        False])
                    co = copy_callback_obj

                    # Save repo path location
                    with open(self.repo_path_saved_file, 'w') as f:
                        f.write(new_repo_path)

                pre, ret, post = self.repository_manager.build_code(
                                                            clean_tmp=True, \
                                                            reconfigure=True, \
                                                            callback_object=co)
                if pre == common_mix.GlobalConstants.COMMAND_FAILURE:
                    ERROR_HANDLER.error_exit("default build failed (pre).", \
                                                                     __file__)
                if ret == common_mix.GlobalConstants.COMMAND_FAILURE:
                    ERROR_HANDLER.error_exit("default build failed (ret).", \
                                                                     __file__)
                if post == common_mix.GlobalConstants.COMMAND_FAILURE:
                    ERROR_HANDLER.error_exit("default build failed (post).", \
                                                                     __file__)
                # XXX: Also already copied
        else:
            ERROR_HANDLER.error_exit("TODO: implement baking relevant files")
    #~ def set_repo_to_build_default(self)
#~ class CodeBuildsFactory()



#
# [LICENCE]
#
"""
This module is used through MetaCriteriaTool class
Which access the relevant testcase tools as specified

The tools are organized by programming language
For each language, there is a folder for each tool, 
named after the tool in lowercase

Each criteria tool package have the following in the __init__.py file:
>>> import <Module>.<class extending BaseCriteriaTool> as CriteriaTool
"""

from __future__ import print_function
import os
import sys
import glob
import logging
import shutil

import muteria.common.fs as common_fs
import muteria.common.matrices as common_matrices
import muteria.common.mix as common_mix

from muteria.drivers import ToolsModulesLoader
from muteria.drivers import DriversUtils

from muteria.drivers.checkpoint_handler import CheckPointHandler

from muteria.drivers.criteria.criteria_info import CriteriaToInfoObject

from muteria.drivers.criteria import CriteriaToolType, TestCriteria

ERROR_HANDLER = common_mix.ErrorHandler

class MetaCriteriaTool(object):
    '''
    '''
    TOOL_OBJ_KEY = "tool_obj"
    TOOL_WORKDIR_KEY = "tool_working_dir"

    @classmethod
    def get_toolnames_by_types_by_criteria_by_language(cls):
        """ get imformation about the plugged-in criteria tool drivers.
            :return: a dict having the form:
                    {
                        language: {
                            criterion: {
                                CriteriaToolType: [
                                    (toolname, is_installed?)
                                ]
                            }
                        }
                    }
        """
        modules_dict = ToolsModulesLoader.get_tools_modules( \
                                            ToolsModulesLoader.CRITERIA_TOOLS)
        res = {}
        for language in modules_dict:
            res[language] = {}
            for toolname in modules_dict[language]:
                for tooltype in CriteriaToolType:
                    tooltype_name = tooltype.get_field_value()
                    try:
                        # the tooltype is returned by config.get_tool_type()
                        CriteriaTool = getattr(\
                                            modules_dict[language][toolname],\
                                                                tooltype_name)
                        if CriteriaTool is None:
                            continue
                        for criterion in CriteriaTool.get_supported_criteria():
                            if criterion not in res[language]:
                                res[language][criterion] = {}
                            if tooltype_name not in res[language][criterion]:
                                res[language][criterion][tooltype_name] = []
                            res[language][criterion][tooltype_name].append(\
                                        (toolname, CriteriaTool.installed()))
                    except AttributeError:
                        ERROR_HANDLER.error_exit("{} {} {} {}".format( \
                                "(REPORT BUG) The criteria tool of type", \
                                tooltype_name,\
                                "is not present for criteria tool", toolname),\
                                                                    __file__)
        return res
    #~ def get_toolnames_by_types_by_criteria_by_language()

    def __init__(self, language, meta_test_generation_obj, \
                        criteria_working_dir, \
                        code_builds_factory, \
                        tools_config_by_criterion_dict):
        # Set Constants
        self.modules_dict = ToolsModulesLoader.get_tools_modules(\
                                            ToolsModulesLoader.CRITERIA_TOOLS)
        
        # Set Direct Arguments Variables
        self.language = language.lower()
        self.meta_test_generation_obj = meta_test_generation_obj
        self.criteria_working_dir = criteria_working_dir
        self.code_builds_factory = code_builds_factory
        self.tools_config_by_criterion_dict = tools_config_by_criterion_dict
        
        # Verify Direct Arguments Variables
        ERROR_HANDLER.assert_true(self.criteria_working_dir is not None, \
                            "Must specify criteria_working_dir", __file__)
        for criterion in self.tools_config_by_criterion_dict:
            ERROR_HANDLER.assert_true( \
                    len(self.tools_config_by_criterion_dict[criterion]) == \
                    len(set([c.get_tool_config_alias() for c in \
                        self.tools_config_by_criterion_dict[criterion]])), \
                    "some tool configs appear multiple times for {}".format( \
                                                        criterion), __file__)
        
        # Set Indirect Arguments Variables
        self.checkpoints_dir = os.path.join(self.criteria_working_dir, \
                                                            "_checkpoints_")
        self.criteria_info_file_by_criteria = {}
        for criterion in self.tools_config_by_criterion_dict:
            self.criteria_info_file_by_criteria[criterion] = os.path.join(\
                                        self.criteria_working_dir, \
                                        criterion.get_str()+"_info_file.json")


        # Verify indirect Arguments Variables

        # Initialize Other Fields
        self.criteria_configured_tools = {}
        self.checkpointer = None
        
        # Make Initialization Computation
        ## Create dirs
        if not os.path.isdir(self.criteria_working_dir):
            self.clear_working_dir()

        ## set checkpointer
        self.checkpointer = common_fs.CheckpointState(\
                                                *self._get_checkpoint_files())

        ## Create the different tools
        ### XXX Merge configs that are identical across criteria
        tmp_alias2conf = {}
        for criterion in self.tools_config_by_criterion_dict:
            for config in self.tools_config_by_criterion_dict[criterion]:
                toolname = config.get_tool_name()
                toolalias = config.get_tool_config_alias()
                if toolalias in tmp_alias2conf:
                    # make sure that the config is the same
                    ERROR_HANDLER.assert_true(config == \
                                                tmp_alias2conf[toolalias], \
                                        "{} {} {}".format( \
                                    "same tool alias but different", \
                                "config for alias", toolalias), __file__)
                else:
                    tmp_alias2conf[toolalias] = config
                    tool_working_dir = self._get_criteria_tool_out_folder(\
                                                                    toolalias)
                    tool_checkpointer = common_fs.CheckpointState( \
                            *self._get_criteria_tool_checkpoint_files( \
                                                                    toolalias))
                    self.checkpointer.add_dep_checkpoint_state( \
                                                            tool_checkpointer)
                    self.criteria_configured_tools[toolalias] = {
                        self.TOOL_OBJ_KEY: self._create_criteria_tool( \
                                                    toolname, \
                                                    tool_working_dir, config, \
                                                    tool_checkpointer),
                        self.TOOL_WORKDIR_KEY: tool_working_dir,
                    }

        # verify supported criteria
        for criterion in self.tools_config_by_criterion_dict:
            for toolconf in self.tools_config_by_criterion_dict[criterion]:
                toolalias = toolconf.get_tool_config_alias()
                if criterion not in self.criteria_configured_tools[toolalias]\
                                [self.TOOL_OBJ_KEY].get_supported_criteria():
                    ERROR_HANDLER.error_exit( \
                            "tool {} specified for unsupported criterion {}".\
                                        format(toolalias, criterion), __file__)
    #~ def __init__()

    def _create_criteria_tool(self, toolname, tool_working_dir, \
                                                    config, tool_checkpointer):
        '''

        '''
        ERROR_HANDLER.assert_true( \
            toolname in self.modules_dict[self.language],
            "Invalid toolname given: {}".format(toolname), __file__)
        try:
            # the tooltype is returned by config.get_tool_type()
            CriteriaTool = getattr( \
                                self.modules_dict[self.language][toolname],\
                                    config.get_tool_type().get_field_value())
        except AttributeError:
            ERROR_HANDLER.error_exit("{} {} {} {}".format( \
                                "(REPORT BUG) The Criteria tool of type", \
                                config.get_tool_type().get_field_value(),\
                                "is not present for test tool", toolname), \
                                                                    __file__)

        ERROR_HANDLER.assert_true(CriteriaTool is not None, \
                                "The {} language's tool {} {} {}.".format( \
                                self.language, toolname, \
                            "does not implement the criteria tool type", \
                                config.get_tool_type().get_str()), __file__)
        crit_tool = CriteriaTool(self.meta_test_generation_obj, \
                                                tool_working_dir, \
                                                self.code_builds_factory, \
                                                    config, tool_checkpointer)
        return crit_tool
    #~ def _create_criteria_tool()

    def check_tools_installed(self):
        non_installed = []
        for toolalias, tool_obj in self.criteria_configured_tools.items():
            if not tool_obj[self.TOOL_OBJ_KEY].tool_installed():
                non_installed.append(toolalias)
        if len(non_installed) > 0:
            ERROR_HANDLER.error_exit("{}: {}".format(\
                            "The following Criteria tools are not installed", \
                            str(non_installed)))
    #~ def check_tools_installed()

    def clear_working_dir(self):
        if os.path.isdir(self.criteria_working_dir):
            shutil.rmtree(self.criteria_working_dir)
        os.mkdir(self.criteria_working_dir)
        if os.path.isdir(self.checkpoints_dir):
            shutil.rmtree(self.checkpoints_dir)
        os.mkdir(self.checkpoints_dir)
        for _, tool_dat in list(self.criteria_configured_tools.items()):
            tool_dat[self.TOOL_OBJ_KEY].clear_working_dir()
    #~ def clear_working_dir()    

    def _get_tool2criteria(self, criteria_passed):
        ''' TODO: update users of this function
            Take a list of criteria and group by tools
        :param criteria_passed: list representing a criteria considered 
        :return: return the criteria by tools
        '''
        tool2criteria = {}
        for criterion in criteria_passed:
            for config in self.tools_config_by_criterion_dict[criterion]:
                ctoolalias = config.get_tool_config_alias()
                if ctoolalias not in tool2criteria:
                    tool2criteria[ctoolalias] = []
                tool2criteria[ctoolalias].append(criterion)
        return tool2criteria
    #~ def _get_tool2criteria()
                                            
    def runtests_criteria_coverage (self, testcases, criterion_to_matrix, \
                                    criterion_to_executionoutput=None,
                                    criteria_element_list_by_criteria=None, \
                                    re_instrument_code=False, \
                                    cover_criteria_elements_once=False,
                                    prioritization_module_by_criteria=None,
                                    parallel_count=1, \
                                    parallel_criteria_test_scheduler=None,\
                                    restart_checkpointer=False, \
                                    finish_destroy_checkpointer=True):
        ''' 
        Executes the instrumented executable code with testscases and
        returns the different code coverage matrices.

        :param testcases: list of testcases to execute

        :param criterion_to_matrix: dict of <criterion, Matrix file 
                        where to store coverage>. 
        :param criterion_to_executionoutput: dict of <criterion, execoutput 
                        file where to store coverage>. 
        
        :param criteria_element_list_by_criteria: dictionary representing the
                        list of criteria elements (stmts, branches, mutants)
                        to consider in the test execution matices. 
                        Key is the criterion and the value the list of elements

        :param re_instrument_code: Decide whether to instrument code before 
                        running the tests. (Example when instrumentation was 
                        not specifically called. This is True by default)

        :param cover_criteria_elements_once: Specify whether to cover criteria
                        elements once is enough, meaning that we stop 
                        analysing a criterion element once a test covers it.
                        The remaining test covering verdict will be UNKNOWN. 

        :param prioritization_module_by_criteria: dict of prioritization module
                        by criteria. None means no prioritization used.

        :type \parallel_count:
        :param \parallel_count:

        :type \parallel_criteria_test_scheduler:
        :param \parallel_criteria_test_scheduler: scheduler that organize 
                        parallelism across criteria tools.
                        (TODO: Implement support)

        :type \restart_checkpointer:
        :param \restart_checkpointer:

        :type finish_destroy_checkpointer:
        :param finish_destroy_checkpointer:
        '''

        # FIXME: Make sure that the support are implemented for 
        # parallelism and test prioritization. Remove the code bellow 
        # once supported:
        ERROR_HANDLER.assert_true(parallel_count <= 1, \
                    "Must implement parallel execution support here", \
                                                                    __file__)
        ERROR_HANDLER.assert_true(parallel_criteria_test_scheduler is None, \
            "Must implement parallel codes tests execution support here", \
                                                                    __file__)
        #~FIXMEnd

        # Check arguments Validity
        ERROR_HANDLER.assert_true(parallel_count > 0, \
                    "invalid parallel  execution count: {}. {}".format( \
                                    parallel_count, "must be >= 1"))

        # @Checkpoint: create a checkpoint handler
        cp_func_name = "runtests_criteria_coverage"
        cp_task_id = 1
        checkpoint_handler = CheckPointHandler( \
                                            self.get_checkpoint_state_object())
        if restart_checkpointer:
            checkpoint_handler.restart()
        if checkpoint_handler.is_finished():
            return

        ERROR_HANDLER.assert_true(len(criterion_to_matrix) > 0, \
                                        "no criterion is enabled", __file__)

        ERROR_HANDLER.assert_true(len(set(criterion_to_matrix) - \
                            set(self.tools_config_by_criterion_dict)) == 0, \
                    "Passed matrices output are more than tool specified", \
                                                                    __file__)

        if criterion_to_executionoutput is not None:
            ERROR_HANDLER.assert_true(set(criterion_to_matrix) == \
                                        set(criterion_to_executionoutput), \
                            "criteria mismatch between matrix and output", \
                                                                    __file__)

        tool2criteria = self._get_tool2criteria(criterion_to_matrix.keys())

        matrices_dir_tmp = os.path.join(self.criteria_working_dir, \
                                                            "codecov_dir.tmp")
        if os.path.isdir(matrices_dir_tmp):
            if restart_checkpointer:
                shutil.rmtree(matrices_dir_tmp)
                os.mkdir(matrices_dir_tmp)
        else:
            os.mkdir(matrices_dir_tmp)

        if criteria_element_list_by_criteria is None:
            criteria_element_list_by_criteria = \
                                        {c: None for c in criterion_to_matrix}

        # get criteria elements by tools
        criteria_elem_list_by_tool = {}
        for criterion in criteria_element_list_by_criteria:
            if criteria_element_list_by_criteria[criterion] is None:
                for t_conf in self.tools_config_by_criterion_dict[criterion]:
                    toolalias = t_conf.get_tool_config_alias()
                    if toolalias not in criteria_elem_list_by_tool:
                        criteria_elem_list_by_tool[toolalias] = {}
                    criteria_elem_list_by_tool[toolalias][criterion] = None
                continue

            criteria_elem_list_by_tool[criterion] = {}
            for crit_elem in criteria_element_list_by_criteria[criterion]:
                toolalias, elem = DriversUtils.reverse_meta_element(crit_elem)
                if toolalias not in criteria_elem_list_by_tool:
                    criteria_elem_list_by_tool[toolalias] = {}
                if criterion not in criteria_elem_list_by_tool[toolalias]:
                    criteria_elem_list_by_tool[toolalias][criterion] = []
                criteria_elem_list_by_tool[toolalias][criterion].append(elem)

            ERROR_HANDLER.assert_true(len(set(criteria_elem_list_by_tool) - \
                                set(self.criteria_configured_tools)) == 0, \
                                "some tool in data not registered", __file__)

        crit2tool2matrixfile = {cv: {} for cv in criterion_to_matrix}
        crit2tool2outhashfile = {cv: {} for cv in criterion_to_executionoutput}
        for ctoolalias in tool2criteria:
            _criteria2matrix = {}
            _criteria2outhash = {}
            for criterion in tool2criteria[ctoolalias]:
                _criteria2matrix[criterion] = os.path.join(matrices_dir_tmp, \
                                                criterion.get_field_value() 
                                                                + '-' 
                                                                + ctoolalias 
                                                                + '.csv')
                if criterion_to_executionoutput is None or \
                            criterion_to_executionoutput[criterion] is None:
                    _criteria2outhash[criterion] = None
                else:
                    _criteria2outhash[criterion] = \
                                            os.path.join(matrices_dir_tmp, \
                                                criterion.get_field_value() 
                                                        + '-' 
                                                        + ctoolalias 
                                                        + '.outloghash.json')
                crit2tool2matrixfile[criterion][ctoolalias] = \
                                                    _criteria2matrix[criterion]
                crit2tool2outhashfile[criterion][ctoolalias] = \
                                                _criteria2outhash[criterion]

            # @Checkpoint: Check whether already executed
            if checkpoint_handler.is_to_execute( \
                                        func_name=cp_func_name, \
                                        taskid=cp_task_id, \
                                        tool=ctoolalias):
                for criterion in _criteria2matrix:
                    _criteria2matrix[criterion] = \
                                        common_matrices.ExecutionMatrix( \
                                        filename=_criteria2matrix[criterion], \
                                        non_key_col_list=testcases)
                    if _criteria2outhash[criterion] is not None:
                        _criteria2outhash[criterion] = \
                                        common_matrices.OutputLogData( \
                                        filename=_criteria2outhash[criterion])
                # Actual execution
                ctool = self.criteria_configured_tools[ctoolalias][\
                                                            self.TOOL_OBJ_KEY]
                ctool.runtests_criteria_coverage(testcases, \
                                criteria_element_list_by_criteria=\
                                        criteria_elem_list_by_tool[toolalias],\
                                criterion_to_matrix=_criteria2matrix, \
                                criterion_to_executionoutput=\
                                                            _criteria2outhash,\
                                re_instrument_code=re_instrument_code, \
                                cover_criteria_elements_once=\
                                                cover_criteria_elements_once, \
                                prioritization_module_by_criteria=\
                                            prioritization_module_by_criteria)

                # Checkpointing
                checkpoint_handler.do_checkpoint( \
                                        func_name=cp_func_name, \
                                        taskid=cp_task_id, \
                                        tool=ctoolalias)

        # Aggregate the matrices and out hashes
        ## Create reult matrices and out hashes
        result_matrices = {}
        result_outloghashes = {}
        for criterion in criterion_to_matrix:
            result_matrices[criterion] = common_matrices.ExecutionMatrix( \
                                filename=criterion_to_matrix[criterion], \
                                non_key_col_list=testcases)
            if criterion_to_executionoutput[criterion] is None:
                result_outloghashes[criterion] = None
            else:
                result_outloghashes[criterion] = \
                            common_matrices.OutputLogData(filename=\
                                    criterion_to_executionoutput[criterion])
                ERROR_HANDLER.assert_true(\
                            crit2tool2outhashfile[criterion] is not None,
                            "Bug: log enabled but hidden from tool", __file__)
        ## Actual aggregate
        for criterion in result_matrices:
            result_matrix = result_matrices[criterion]
            result_outloghash = result_outloghashes[criterion]
            for mtoolalias in crit2tool2matrixfile[criterion]:
                tool_matrix = common_matrices.ExecutionMatrix(\
                        filename=crit2tool2matrixfile[criterion][mtoolalias])

                # Check columns
                ERROR_HANDLER.assert_true(tool_matrix.get_key_colname() == \
                                            result_matrix.get_key_colname(), \
                                    "mismatch on key column name", __file__)
                ERROR_HANDLER.assert_true( \
                                set(tool_matrix.get_nonkey_colname_list()) == \
                                set(result_matrix.get_nonkey_colname_list()), \
                                "mismatch on non key column names", __file__)

                # bring in the data
                key2nonkeydict = tool_matrix.to_pandas_df().\
                        set_index(tool_matrix.get_key_colname(), drop=True).\
                                                to_dict(orient="index")
                
                for c_key in key2nonkeydict:
                    meta_c_key = DriversUtils.make_meta_element(\
                                                        str(c_key), mtoolalias)
                    result_matrix.add_row_by_key(meta_c_key, 
                                                        key2nonkeydict[c_key], 
                                                        serialize=False)

                # out log hash
                if crit2tool2outhashfile[criterion] is not None:
                    tool_outloghash = common_matrices.OutputLogData(\
                            filename=\
                                crit2tool2outhashfile[criterion][mtoolalias])
                    for objective, objective_data in \
                                tool_outloghash.get_zip_objective_and_data():
                        meta_objective = DriversUtils.make_meta_element(\
                                                    str(objective), mtoolalias)
                        result_outloghash.add_data(
                                            {meta_objective: objective_data}, \
                                            serialize=False)

            # @Checkpoint: Check whether already executed
            if checkpoint_handler.is_to_execute( \
                                        func_name=cp_func_name, \
                                        taskid=cp_task_id + 1,
                                        tool=criterion.get_str()):
                # Serialized the computed matrix
                result_matrix.serialize()
                if result_outloghash is not None:
                    result_outloghash.serialize()
            # @Checkpoint: Checkpointing
            checkpoint_handler.do_checkpoint( \
                                    func_name=cp_func_name, \
                                    taskid=cp_task_id + 1,
                                    tool=criterion.get_str())
        
        # Delete the temporary tool matrix's directory
        if os.path.isdir(matrices_dir_tmp):
            shutil.rmtree(matrices_dir_tmp)

        # @Checkpoint: Finished
        detailed_exectime = {}
        for ctoolalias in tool2criteria:
            ct = self.criteria_configured_tools[ctoolalias][self.TOOL_OBJ_KEY]
            detailed_exectime[ctoolalias] = (\
                        ct.get_checkpointer().get_execution_time(),\
                        ct.get_checkpointer().get_detailed_execution_time())

        checkpoint_handler.set_finished(\
                                    detailed_exectime_obj=detailed_exectime)

        if finish_destroy_checkpointer:
            checkpoint_handler.destroy()
    #~ def runtests_criteria_coverage()

    def instrument_code (self, criteria_enabled_list=None, \
                    exe_path_map=None, \
                    #outputdir_override=None, \
                    #code_builds_factory_override=None, \
                    parallel_count=1, \
                    restart_checkpointer=False, \
                    finish_destroy_checkpointer=True):
        """ Instrument the code for the criteria measurements. 

        :type criteria_enabled_list: dict or None
        :param criteria_enabled_list: When None, use all supported criteria
                    else use the specified criteria
    
        :type \exe_path_map: dict or None
        :param \exe_path_map: When None, use all exe, else instrument 
                    files as dict key and write the instrumented output
                    in directory as value. 
    
        :type \parallel_count:
        :param \parallel_count:
    
        :type \restart_checkpointer:
        :param \restart_checkpointer:
    
        :type \finish_destroy_checkpointer:
        :param \finish_destroy_checkpointer:
    
        :raises:
    
        :rtype:
        """
        # FIXME: Support parallelism, then remove the code
        # bellow:
        ERROR_HANDLER.assert_true(parallel_count <= 1, \
                "FIXME: Must first implement support for parallel")
        #~ FXIMEnd

        # Check arguments Validity
        ERROR_HANDLER.assert_true(parallel_count > 0, \
                    "invalid parallel  execution count: {}. {}".format( \
                            parallel_count, "must be >= 1"), __file__)

        # @Checkpoint: create a checkpoint handler
        cp_func_name = "instrument_code"
        cp_task_id = 1
        checkpoint_handler = CheckPointHandler( \
                                            self.get_checkpoint_state_object())
        if restart_checkpointer:
            checkpoint_handler.restart()
        if checkpoint_handler.is_finished():
            return

        if criteria_enabled_list is None:
            criteria_enabled_list = self.tools_config_by_criterion_dict.keys()
        else:
            ERROR_HANDLER.assert_true(len(criteria_enabled_list) > 0, \
                                        "no criterion is enabled", __file__)

            ERROR_HANDLER.assert_true(len(set(criteria_enabled_list) - \
                            set(self.tools_config_by_criterion_dict)) == 0, \
                        "Passed matrice output are more than toll specified", \
                                                                    __file__)

        tool2criteria = self._get_tool2criteria(criteria_enabled_list)

        for ctoolalias in tool2criteria:
            # @Checkpoint: Check whether already executed
            if checkpoint_handler.is_to_execute( \
                                        func_name=cp_func_name, \
                                        taskid=cp_task_id, \
                                        tool=ctoolalias):
                # Actual execution
                ctool = self.criteria_configured_tools[ctoolalias][\
                                                            self.TOOL_OBJ_KEY]
                ctool.instrument_code(\
                                enabled_criteria=tool2criteria[ctoolalias],\
                                exe_path_map=exe_path_map)
                # ensure repo is set back
                self.code_builds_factory.set_repo_to_build_default()

                # @Checkpoint: Checkpointing
                checkpoint_handler.do_checkpoint( \
                                        func_name=cp_func_name, \
                                        taskid=cp_task_id, \
                                        tool=ctoolalias)

        # Invalidate any existing mutant info so it can be recomputed
        self._invalidate_criteria_info(
                                enabled_criteria=tool2criteria[ctoolalias])
                                

        # @Checkpoint: Finished
        detailed_exectime = {}
        for ctoolalias in tool2criteria:
            ct = self.criteria_configured_tools[ctoolalias][self.TOOL_OBJ_KEY]
            detailed_exectime[ctoolalias] = (\
                        ct.get_checkpointer().get_execution_time(),\
                        ct.get_checkpointer().get_detailed_execution_time())

        checkpoint_handler.set_finished( \
                                    detailed_exectime_obj=detailed_exectime)

        if finish_destroy_checkpointer:
            checkpoint_handler.destroy()
    #~ def instrument_code()

    def _compute_criterion_info(self, criterion, candidate_tool_aliases=None):
        if criterion not in CriteriaToInfoObject:
            return None

        meta_criterion_info_obj = CriteriaToInfoObject[criterion]()
        if candidate_tool_aliases is None:
            candidate_tool_aliases = []
            for config in self.tools_config_by_criterion_dict[criterion]:
                candidate_tool_aliases.append(config.get_tool_config_alias())
        for ctoolalias in candidate_tool_aliases:
            ctool = self.criteria_configured_tools[ctoolalias]\
                                                            [self.TOOL_OBJ_KEY]
            tool_element_info = ctool.get_criterion_info_object(criterion)
            if tool_element_info is not None:
                old2new_tests = {}
                for c_elem in tool_element_info.get_elements_list():
                    meta_c_key = DriversUtils.make_meta_element(\
                                                            c_elem, ctoolalias)
                    old2new_tests[c_elem] = meta_c_key
                meta_criterion_info_obj.update_using(\
                                                ctoolalias, old2new_tests, \
                                                            tool_element_info)
        return meta_criterion_info_obj
    #~ def _compute_criterion_info()
    
    def get_criterion_info_object(self, criterion):
        if criterion not in CriteriaToInfoObject:
            return None
            
        return CriteriaToInfoObject[criterion]().load_from_file(\
                                    self.get_criterion_info_file(criterion))
    #~ def def get_criterion_info_object()

    def get_criterion_info_file(self, criterion):
        if criterion not in CriteriaToInfoObject:
            return None
            
        # Compute and write the testcase info if not present
        # only place where the meta info is written
        info_file = self._unchecked_get_criterion_info_file(criterion)
        if self._criterion_info_is_invalidated(criterion):
            self._compute_criterion_info(criterion).write_to_file(info_file)
        return info_file
    #~ def get_criterion_info_file()

    def _unchecked_get_criterion_info_file(self, criterion):
        if criterion not in CriteriaToInfoObject:
            return None
            
        return self.criteria_info_file_by_criteria[criterion] 
    #~ def _unchecked_get_criteria_info_file():

    def _invalidate_criterion_info(self, criterion):
        if criterion in CriteriaToInfoObject:
            info_file = self._unchecked_get_criterion_info_file(criterion)
            if os.path.isfile(info_file):
                os.remove(info_file)
    #~ def _invalidate_criterion_info()

    def _invalidate_criteria_info(self, enabled_criteria=None):
        if enabled_criteria is None:
            enabled_criteria = self.tools_config_by_criterion_dict.keys()
        for criterion in enabled_criteria:
            if criterion in CriteriaToInfoObject:
                self._invalidate_criterion_info(criterion)
    #~ def _invalidate_criteria_info()

    def _criterion_info_is_invalidated(self, criterion):
        ERROR_HANDLER.assert_true(criterion in CriteriaToInfoObject,
                            "The criterion {} {}".format(criterion.get_str(), \
                            "must have info Class registered for this"), \
                                                                    __file__)
            
        return not os.path.isfile(\
                            self._unchecked_get_criterion_info_file(criterion))
    #~ def _criterion_info_is_invalidated()

    def _get_criteria_tool_out_folder(self, ccov_toolname, top_outdir=None):
        if top_outdir is None:
            top_outdir = self.criteria_working_dir
        return os.path.join(top_outdir, ccov_toolname)
    #~def _get_criteria_tool_out_folder()

    def _get_criteria_tool_checkpoint_files(self, criteria_toolname, \
                                                    top_checkpoint_dir=None):
        if top_checkpoint_dir is None:
            top_checkpoint_dir = self.checkpoints_dir
        return [os.path.join(top_checkpoint_dir, \
                        criteria_toolname+"_checkpoint.state"+suffix) \
                        for suffix in ("", ".backup")]
    #~def _get_criteria_tool_checkpoint_files()
        
    def _get_checkpoint_files(self):
        top_checkpoint_dir = self.checkpoints_dir
        return [os.path.join(top_checkpoint_dir, \
                        "checkpoint.state"+suffix) \
                        for suffix in ("", ".backup")]
    #~def _get_checkpoint_files()

    def get_criteria_tools_by_name(self, toolname):
        res = {}
        for alias, data in self.criteria_configured_tools.items():
            if data[self.TOOL_OBJ_KEY].get_toolname() == toolname:
                res[alias] = data[self.TOOL_OBJ_KEY]
        return res
    #~ def get_criteria_tools_by_name()

    def get_checkpoint_state_object(self):
        return self.checkpointer
    #~ def get_checkpoint_state_object()

    def has_checkpointer(self):
        return self.checkpointer is not None
    #~ def has_checkpointer()
#~ class MetaCriteriaTool



""" This Module implement the data structore and format for the 
    Checkpointing of the main controller during the execution phase

    The importants classes or elements are:
    - `Tasks` enum: enum that defines the differents tasks that are 
                    checkpointed during the controller execution.    
    - `Status` enum: enum defining the different states that a task can 
                    have (untouched, executing or done)
    - `TaskOrderingDependency` class: That defines the dependency between
                    the different tasks as well as their status.
                    The state changes as well as the task to run next are 
                    also implemented in the class.
"""


from __future__ import print_function

import logging
import collections

import networkx

import muteria.common.mix as common_mix

ERROR_HANDLER = common_mix.ErrorHandler


class Tasks(common_mix.EnumAutoName):
    STARTING = 0 #enum.auto()

    TESTS_GENERATION_GUIDANCE = 1 #enum.auto()
    TESTS_GENERATION = 2 #enum.auto()
    TESTS_GENERATION_USING_CRITERIA = 3 #enum.auto()
    TESTS_EXECUTION_SELECTION_PRIORITIZATION = 4 #enum.auto()
    PASS_FAIL_TESTS_EXECUTION = 5 #enum.auto()

    CRITERIA_GENERATION_GUIDANCE = 6 #enum.auto()
    CRITERIA_GENERATION = 7 #enum.auto()
    CRITERIA_EXECUTION_SELECTION_PRIORITIZATION = 8 #enum.auto()
    CRITERIA_TESTS_EXECUTION = 9 #enum.auto()

    PASS_FAIL_STATS = 10 #enum.auto()
    CRITERIA_STATS = 11 #enum.auto()
    AGGREGATED_STATS = 12 #enum.auto()

    FINISHED = 13 #enum.auto()
#~ class Tasks

class Status(common_mix.EnumAutoName):
    UNTOUCHED = -1 #enum.auto()
    EXECUTING = 0 #enum.auto()
    DONE = 1 #enum.auto()
#~ class Status

class TaskOrderingDependency(object): 
    '''
    The task dependency structure is following(The structure neve change):
    ----------------------------------------------------------------------
        TESTS_GENERATION_GUIDANCE --> STARTING
        CRITERIA_GENERATION_GUIDANCE --> STARTING

        TESTS_GENERATION --> TESTS_GENERATION_GUIDANCE

                                            | TESTS_GENERATION_GUIDANCE
        TESTS_GENERATION_USING_CRITERIA --> | CRITERIA_GENERATION 

                                                     | TESTS_GENERATION_USING_CRITERIA 
        TESTS_EXECUTION_SELECTION_PRIORITIZATION --> | TESTS_GENERATION

        PASS_FAIL_TESTS_EXECUTION --> TESTS_EXECUTION_SELECTION_PRIORITIZATION

        CRITERIA_GENERATION --> CRITERIA_GENERATION_GUIDANCE

        CRITERIA_EXECUTION_SELECTION_PRIORITIZATION --> CRITERIA_GENERATION 

                                     | CRITERIA_EXECUTION_SELECTION_PRIORITIZATION
        CRITERIA_TESTS_EXECUTION --> | PASS_FAIL_TESTS_EXECUTION

        PASS_FAIL_STATS --> PASS_FAIL_TESTS_EXECUTION

        CRITERIA_STATS --> CRITERIA_TESTS_EXECUTION

                             | PASS_FAIL_STATS
        AGGREGATED_STATS --> | CRITERIA_STATS

        FINISHED --> AGGREGATED_STATS
    ------------------------------------------------------------------------

    '''
    def __init__(self, json_obj=None):
        if json_obj is None:
            self.initialize_data_graph()
        else:
            # Check that every task is present
            if type(json_obj) != dict:
                ERROR_HANDLER.error_exit("{} {}".format(\
                        "Invalid object to initialize", \
                        "TaskOrderingDependency, must be a 'dict'"), __file__)
            if len(json_obj) != len(Tasks):
                ERROR_HANDLER.error_exit("{} {}".format(\
                            "Invalid object to initialize", \
                            "TaskOrderingDependency. Size mismatch"), __file__)
            # verify
            for key_t in Tasks:
                if key_t.get_str() not in json_obj:
                    ERROR_HANDLER.error_exit("{} {} {} {}".format( \
                        "Invalid object to initialize TaskOrderingDependency",\
                        "The task", key_t.get_str(), "is absent in json_obj"),\
                                                                    __file__)
                if not Status.has_element_named(json_obj[key_t.get_str()]):
                    ERROR_HANDLER.error_exit("{} {} {} {}".format( \
                        "Invalid object to initialize TaskOrderingDependency",\
                        "The task", key_t.get_str(), "has invalid status"), \
                                                                    __file__)
            # initialize
            decoded_json_obj = {}
            for k_str, v_str in json_obj.items():
                decoded_json_obj[Tasks[k_str]] = Status[v_str]
            self.initialize_data_graph(decoded_json_obj)
    #~ def __init__()

    def get_as_json_object(self):
        ret_obj = {}
        for key_t in Tasks:
            ret_obj[key_t.get_str()] = \
                        self._lookup_task_cell(key_t).get_status().get_str()
        return ret_obj
    #~ def get_as_json_object()

    def export_graph(self, graph_out_filename):
        graph = networkx.Graph()
        # add nodes
        for key_t in Tasks:
            graph.add_node(key_t.get_str())
        # add edges
        visited = set()
        queue = collections.deque()
        visited.add(self.root)
        queue.append(self.root)
        while queue:
            cur_node = queue.popleft()
            for d in cur_node.get_dependencies():
                graph.add_edge(d.get_task_name().get_str(), \
                                        cur_node.get_task_name().get_str())
                if d not in visited:
                    visited.add(d)
                    queue.append(d)
        # print graph
        d_graph = networkx.DiGraph(graph) 
        networkx.write_graphml(d_graph, graph_out_filename)
    #~ def export_graph()

    def initialize_data_graph(self, task_status_map=None):
        '''
        Each param is the status of the corresponding task 
        '''
        # Default behaviour (initial initialization)
        if task_status_map is None:
            task_status_map = {}
            for key_t in Tasks:
                task_status_map[key_t] = Status.UNTOUCHED

        # set the root cell (The root has no uses)
        finished = self.Cell(Tasks.FINISHED)
        finished.set_status(task_status_map[Tasks.FINISHED])

        self.root = finished

        # Construct the rest of the dependence graph, starting at the root
        ## Stats Layer
        ### Aggregate stats
        agg_stats =  self.Cell(Tasks.AGGREGATED_STATS)
        agg_stats.set_status(task_status_map[Tasks.AGGREGATED_STATS]) 
        finished.add_dependency(agg_stats)

        ### Other stats
        passfail_stats = self.Cell(Tasks.PASS_FAIL_STATS)
        passfail_stats.set_status(task_status_map[Tasks.PASS_FAIL_STATS])
        agg_stats.add_dependency(passfail_stats)

        crit_stats = self.Cell(Tasks.CRITERIA_STATS)
        crit_stats.set_status(task_status_map[Tasks.CRITERIA_STATS])
        agg_stats.add_dependency(crit_stats)

        ## Matrices Layer
        crit = self.Cell(Tasks.CRITERIA_TESTS_EXECUTION)
        crit.set_status(task_status_map[Tasks.CRITERIA_TESTS_EXECUTION])
        crit_stats.add_dependency(crit)

        passfail = self.Cell(Tasks.PASS_FAIL_TESTS_EXECUTION)
        passfail.set_status(task_status_map[Tasks.PASS_FAIL_TESTS_EXECUTION])
        passfail_stats.add_dependency(passfail)
        crit.add_dependency(passfail)

        # Artifact Execution Selection and Prioritization
        ce_sp = self.Cell(Tasks.CRITERIA_EXECUTION_SELECTION_PRIORITIZATION)
        ce_sp.set_status(task_status_map[ \
                            Tasks.CRITERIA_EXECUTION_SELECTION_PRIORITIZATION])
        crit.add_dependency(ce_sp)

        t_sp = self.Cell(Tasks.TESTS_EXECUTION_SELECTION_PRIORITIZATION)
        t_sp.set_status(task_status_map[\
                            Tasks.TESTS_EXECUTION_SELECTION_PRIORITIZATION])
        passfail.add_dependency(t_sp)

        # Artifact Generation
        t_gen_crit = self.Cell(Tasks.TESTS_GENERATION_USING_CRITERIA)
        t_gen_crit.set_status(task_status_map[Tasks.TESTS_GENERATION_USING_CRITERIA])
        t_sp.add_dependency(t_gen_crit)

        crit_gen = self.Cell(Tasks.CRITERIA_GENERATION)
        crit_gen.set_status(task_status_map[Tasks.CRITERIA_GENERATION])
        ce_sp.add_dependency(crit_gen)
        t_gen_crit.add_dependency(crit_gen)

        t_gen = self.Cell(Tasks.TESTS_GENERATION)
        t_gen.set_status(task_status_map[Tasks.TESTS_GENERATION])
        t_sp.add_dependency(t_gen)

        t_gguide = self.Cell(Tasks.TESTS_GENERATION_GUIDANCE)
        t_gguide.set_status(task_status_map[Tasks.TESTS_GENERATION_GUIDANCE])
        t_gen.add_dependency(t_gguide)
        t_gen_crit.add_dependency(t_gguide)

        c_gguide = self.Cell(Tasks.CRITERIA_GENERATION_GUIDANCE)
        c_gguide.set_status(task_status_map[Tasks.CRITERIA_GENERATION_GUIDANCE])
        crit_gen.add_dependency(c_gguide)

        # Starting
        starting = self.Cell(Tasks.STARTING)
        starting.set_status(task_status_map[Tasks.STARTING])
        c_gguide.add_dependency(starting)
        t_gguide.add_dependency(starting)

        # COMPUTE THE USES AND VERIFY
        visited = set()
        self._compute_uses_recursive(self.root, visited)

        ## Verify
        number_of_tasks = len(Tasks) # Number of instances in 'Tasks' class
        if len(visited) != number_of_tasks:
            ERROR_HANDLER.error_exit("%s %s" % (\
                    "BUG in the Task Ordering Dependency. got", \
                    "%d but expects %d" % (len(visited), number_of_tasks)),\
                                                                    __file__)
        visited = {}
        self._recursive_verify(self.root, visited)
    #~ def initialize_data_graph()

    def set_task_completed (self, task_name):
        # look the task up from the root
        t = self._lookup_task_cell(task_name)
        # Check that all deps are done
        self._recursive_check_deps_are_done(t)
        # set it to complete
        t.set_done()
        logging.info("checkpoint_tasks:COMPLETED:"+task_name.get_str())
    #~ def set_task_completed()

    def set_task_executing (self, task_name):
        # look the task up from the root
        t = self._lookup_task_cell(task_name)
        # Check that all deps are done
        self._recursive_check_deps_are_done(t)
        # set it to complete
        t.set_executing()
        logging.info("checkpoint_tasks:EXECUTING:"+task_name.get_str())
    #~ def set_task_executing()

    def task_is_complete(self, task_name):
        t = self._lookup_task_cell(task_name)
        return t.is_done()
    #~ def task_is_complete()

    def task_is_executing(self, task_name):
        t = self._lookup_task_cell(task_name)
        return t.is_executing()
    #~ def task_is_executing()

    def task_is_untouched(self, task_name):
        t = self._lookup_task_cell(task_name)
        return t.is_untouched()
    #~ def task_is_untouched()

    def get_next_todo_tasks(self):
        task_node_set = set()
        self._recursive_get_next_todo_tasks(task_node_set, self.root)
        task_set = {x.get_task_name() for x in task_node_set}
        if len(task_set) != len(task_node_set):
            ERROR_HANDLER.error_exit("{}".format(\
                        "(BUG) same task appears in multiple nodes"), __file__)
        logging.info("checkpoint_tasks:NEXT_TODO:"+\
                                        str([t.get_str() for t in task_set]))
        return task_set
    #~ def get_next_todo_tasks()

    def set_all_tasks_to_completed(self):
        task_name_set = self.get_next_todo_tasks()
        for task_name in task_name_set:
            t = self._lookup_task_cell(task_name)
            self._recursive_set_task_as_done(t)
    #~ def set_all_tasks_to_completed()

    def set_task_back_as_todo_executing(self, task_name):
        t = self._lookup_task_cell(task_name)
        if not t.is_executing():
            t.set_executing()
            for u in t.get_uses():
                self._recursive_set_task_back_as_todo_untouched(u)
    #~ def set_task_back_as_todo_executing()

    def set_task_back_as_todo_untouched(self, task_name):
        t = self._lookup_task_cell(task_name)
        self._recursive_set_task_back_as_todo_untouched(t)  
    #~ def set_task_back_as_todo_untouched()

    # ==== PRIVATE ====

    class Cell():
        def __init__(self, task_name, status=Status.UNTOUCHED):
            self.dependencies = set()
            self.uses = set()
            self.status = status
            self.task_name = task_name
            if not Status.is_valid(self.status):
                ERROR_HANDLER.error_exit(\
                            "Invalid value passed as status, use 'Status.'", \
                                                                    __file__)

        # setters
        def add_dependency(self, cell):
            self.dependencies.add(cell)
        def add_use(self, cell):
            self.uses.add(cell)
        def set_status(self, status):
            if not Status.is_valid(status):
                ERROR_HANDLER.error_exit( \
                    "Invalid status value passed in set_status.use 'Status.'",\
                                                                    __file__)
            self.status = status
        def set_done(self):
            self.status = Status.DONE
        def set_executing(self):
            self.status = Status.EXECUTING
        def set_untouched(self):
            self.status = Status.UNTOUCHED
        # getters
        def get_dependencies(self):
            return self.dependencies
        def get_uses(self):
            return self.uses
        def get_status(self):
            return self.status
        def get_task_name(self):
            return self.task_name
        # extra getters
        def is_done(self):
            return self.status == Status.DONE
        def is_executing(self):
            return self.status == Status.EXECUTING
        def is_untouched(self):
            return self.status == Status.UNTOUCHED
    #~ class Cell

    def _compute_uses_recursive(self, start_node, visited):
        '''
        Recursively set the use of the nodes (in the dependencies)
        Using DFS, just update the already visited deps, no recursive call

        :param start_node: currently explored node
        :param visited: set of visited nodes (already called recusively)
        '''
        if start_node not in visited:
            visited.add(start_node)
            for d in start_node.get_dependencies():
                d.add_use(start_node)
                self._compute_uses_recursive(d, visited)
    #~ def _compute_uses_recursive() 

    def _recursive_verify(self, start_node, visited):
        # verify no loop and status soundness
        if start_node in visited:
            ERROR_HANDLER.error_exit(\
                            "There is a loop involving task {}".format( \
                            start_node.get_task_name().get_str()), __file__)
        for d in start_node.get_dependencies():
            if not start_node.is_untouched():
                if not d.is_done():
                    ERROR_HANDLER.error_exit("{} {} {} {} {}".format( \
                                    "Unsoundness of status. dependency", \
                                    d.get_task_name().get_str(), \
                                    "not done while use", \
                                    start_node.get_task_name().get_str(), \
                                    "is either executing or done"), __file__)
            self._recursive_verify(d, visited.copy())
    #~ def _recursive_verify()

    def _recursive_lookup(self, task_name, start_node):
        if start_node.task_name == task_name:
            return start_node
        for d_node in start_node.get_dependencies():
            ret = self._recursive_lookup(task_name, d_node)
            if ret is not None:
                return ret
        return None
    #~ def _recursive_lookup()

    def _lookup_task_cell(self, task_name):
        cell = self._recursive_lookup(task_name, self.root)
        if cell is None:
            ERROR_HANDLER.error_exit("%s %s" % \
                ("The Task looked up does not exist in dep graph", task_name),\
                                                                    __file__)
        return cell
    #~ def _lookup_task_cell()

    def _recursive_check_deps_are_done(self, start_node):
        problem = None
        for d in start_node.get_dependencies():
            if not d.is_done():
                problem = d.get_task_name
                break
            self._recursive_check_deps_are_done(d)
        if problem is not None:
            ERROR_HANDLER.error_exit("%s %s %s %s" % (\
                        "task", start_node.get_task_name, \
                        "is done while the dependencies are not:", problem),\
                                                                    __file__)
    #~ def _recursive_check_deps_are_done()

    def _recursive_set_task_as_done(self, start_node):
        if not start_node.is_done():
            start_node.set_done()
            for u in start_node.get_uses():
                self._recursive_set_task_as_done(u)
    #~ def _recursive_set_task_as_done()

    def _recursive_set_task_back_as_todo_executing(self, start_node):
        if not start_node.is_executing():
            start_node.set_executing()
            for u in start_node.get_uses():
                self._recursive_set_task_back_as_todo_executing(u)
    #~ def _recursive_set_task_back_as_todo_executing()

    def _recursive_set_task_back_as_todo_untouched(self, start_node):
        if not start_node.is_untouched():
            start_node.set_untouched()
            for u in start_node.get_uses():
                self._recursive_set_task_back_as_todo_untouched(u)
    #~ def _recursive_set_task_back_as_todo_untouched()

    def _recursive_get_next_todo_tasks(self, task_set, start_node):
        if start_node.is_done():
            return
        is_leaf = True
        for d in start_node.get_dependencies():
            if not d.is_done():
                is_leaf = False
                self._recursive_get_next_todo_tasks(task_set, d)
        if is_leaf:
            task_set.add(start_node)
    #~ def _recursive_get_next_todo_tasks()
#~ class TaskOrderingDependency


